# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
import warnings
import requests
import pandas as pd

from doc_utils import DocUtils
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union
from tqdm.auto import tqdm

from relevanceai._api import APIClient

from relevanceai.utils.logger import FileLogger
from relevanceai.utils.decorators.analytics import track
from relevanceai.utils import make_id

from relevanceai.constants.warning import Warning


class Write(APIClient):
    @track
    def insert_documents(
        self,
        documents: list,
        bulk_fn: Callable = None,
        max_workers: int = 8,
        retry_chunk_mult: float = 0.5,
        show_progress_bar: bool = False,
        chunksize: int = 0,
        use_json_encoder: bool = True,
        create_id: bool = False,
        **kwargs,
    ) -> Dict:

        """
        Insert a list of documents with multi-threading automatically enabled.

        - When inserting the document you can optionally specify your own id for a document by using the field name "_id", if not specified a random id is assigned.
        - When inserting or specifying vectors in a document use the suffix (ends with) "_vector_" for the field name. e.g. "product_description_vector_".
        - When inserting or specifying chunks in a document the suffix (ends with) "_chunk_" for the field name. e.g. "products_chunk_".
        - When inserting or specifying chunk vectors in a document's chunks use the suffix (ends with) "_chunkvector_" for the field name. e.g. "products_chunk_.product_description_chunkvector_".

        Documentation can be found here: https://ingest-api-dev-aueast.relevance.ai/latest/documentation#operation/InsertEncode

        Parameters
        ----------
        documents: list
            A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '_id', for specifying vector field use the suffix of '_vector_'
        bulk_fn : callable
            Function to apply to documents before uploading
        max_workers : int
            Number of workers active for multi-threading
        retry_chunk_mult: int
            Multiplier to apply to chunksize if upload fails
        chunksize : int
            Number of documents to upload per worker. If None, it will default to the size specified in config.upload.target_chunk_mb
        use_json_encoder : bool
            Whether to automatically convert documents to json encodable format

        Example
        --------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            documents = [
                {
                    "_id": "10",
                    "value": 5
                },
                {
                    "_id": "332",
                    "value": 10
                }
            ]

            df.insert_documents(documents)

        """
        results = self._insert_documents(
            dataset_id=self.dataset_id,
            documents=documents,
            bulk_fn=bulk_fn,
            max_workers=max_workers,
            retry_chunk_mult=retry_chunk_mult,
            show_progress_bar=show_progress_bar,
            chunksize=chunksize,
            use_json_encoder=use_json_encoder,
            create_id=create_id,
            **kwargs,
        )
        return self._process_insert_results(results)

    @track
    def insert_csv(
        self,
        filepath_or_buffer,
        chunksize: int = 10000,
        max_workers: int = 8,
        retry_chunk_mult: float = 0.5,
        show_progress_bar: bool = False,
        index_col: int = None,
        csv_args: Optional[dict] = None,
        col_for_id: str = None,
        auto_generate_id: bool = True,
    ) -> Dict:
        """
        Insert data from csv file

        Parameters
        ----------
        filepath_or_buffer :
            Any valid string path is acceptable. The string could be a URL. Valid URL schemes include http, ftp, s3, gs, and file.
        chunksize : int
            Number of lines to read from csv per iteration
        max_workers : int
            Number of workers active for multi-threading
        retry_chunk_mult: int
            Multiplier to apply to chunksize if upload fails
        csv_args : dict
            Optional arguments to use when reading in csv. For more info, see https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
        index_col : None
            Optional argument to specify if there is an index column to be skipped (e.g. index_col = 0)
        col_for_id : str
            Optional argument to use when a specific field is supposed to be used as the unique identifier ('_id')
        auto_generate_id: bool = True
            Automatically generateds UUID if auto_generate_id is True and if the '_id' field does not exist

        Example
        ---------
        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")

            csv_filename = "temp.csv"
            df.insert_csv(csv_filename)

        """
        csv_args = {} if csv_args is None else csv_args

        results = self._insert_csv(
            dataset_id=self.dataset_id,
            filepath_or_buffer=filepath_or_buffer,
            chunksize=chunksize,
            max_workers=max_workers,
            retry_chunk_mult=retry_chunk_mult,
            show_progress_bar=show_progress_bar,
            index_col=index_col,
            csv_args=csv_args,
            col_for_id=col_for_id,
            auto_generate_id=auto_generate_id,
        )
        self._process_insert_results(results)
        return results

    @track
    def insert_pandas_dataframe(
        self, df: pd.DataFrame, col_for_id=None, *args, **kwargs
    ):
        """
        Insert a dataframe into the dataset.
        Takes additional args and kwargs based on `insert_documents`.

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")
            pandas_df = pd.DataFrame({"value": [3, 2, 1], "_id": ["10", "11", "12"]})
            df.insert_pandas_dataframe(pandas_df)

        """
        if col_for_id is not None:
            df["_id"] = df[col_for_id]

        else:
            uuids = [make_id(df.iloc[index]) for index in range(len(df))]
            df["_id"] = uuids

        def _is_valid(v):
            try:
                if pd.isna(v):
                    return False
                else:
                    return True
            except:
                return True

        documents = [
            {k: v for k, v in doc.items() if _is_valid(v)}
            for doc in df.to_dict(orient="records")
        ]

        results = self._insert_documents(self.dataset_id, documents, *args, **kwargs)
        self.print_search_dashboard_url(self.dataset_id)
        return results

    def insert_media_folder(
        self,
        path: Union[Path, str],
        field: str = "medias",
        recurse: bool = True,
        *args,
        **kwargs,
    ):
        """
        Given a path to a directory, this method loads all media-related files
        into a Dataset.

        Parameters
        ----------
        field: str
            A text field of a dataset.

        path: Union[Path, str]
            The path to the directory containing medias.

        recurse: bool
            Indicator that determines whether to recursively insert medias from
            subdirectories in the directory.

        Returns
        -------
            dict

        Example
        -------
        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("dataset_id")

            from pathlib import Path
            path = Path("medias/")
            # list(path.iterdir()) returns
            # [
            #    PosixPath('media.jpg'),
            #    PosixPath('more-medias'), # a directory
            # ]

            get_all_medias: bool = True
            if get_all_medias:
                # Inserts all medias, even those in the more-medias directory
                ds.insert_media_folder(
                    field="medias", path=path, recurse=True
                )
            else:
                # Only inserts media.jpg
                ds.insert_media_folder(
                    field="medias", path=path, recurse=False
                )

        """
        if isinstance(path, str):
            path = Path(path)

            if not path.is_dir():
                raise Exception(f"{path} is not a proper path")

        from mimetypes import types_map

        media_extensions = set(
            k.lower() for k, v in types_map.items() if v.startswith("media/")
        )

        def get_paths(path: Path, medias: List[str]) -> List[str]:
            for file in path.iterdir():
                if file.is_dir() and recurse:
                    medias.extend(get_paths(file, []))
                elif file.is_file() and file.suffix.lower() in media_extensions:
                    medias.append(str(file))
                else:
                    continue

            return medias

        medias = get_paths(path, [])
        documents = list(
            map(
                lambda media: {"_id": make_id(media), "path": media, field: media},
                medias,
            )
        )
        results = self.insert_documents(documents, *args, **kwargs)
        self.image_fields.append(field)
        return results

    @track
    def upsert_documents(
        self,
        documents: list,
        bulk_fn: Callable = None,
        max_workers: int = 8,
        retry_chunk_mult: float = 0.5,
        chunksize: int = 0,
        show_progress_bar=False,
        use_json_encoder: bool = True,
        return_json: bool = False,
        create_id: bool = False,
    ) -> Dict:

        """
        Update a list of documents with multi-threading automatically enabled.
        Edits documents by providing a key value pair of fields you are adding or changing, make sure to include the "_id" in the documents.


        Parameters
        ----------
        documents : list
            A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '_id', for specifying vector field use the suffix of '_vector_'
        bulk_fn : callable
            Function to apply to documents before uploading
        max_workers : int
            Number of workers active for multi-threading
        retry_chunk_mult: int
            Multiplier to apply to chunksize if upload fails
        chunksize : int
            Number of documents to upload per worker. If None, it will default to the size specified in config.upload.target_chunk_mb
        use_json_encoder : bool
            Whether to automatically convert documents to json encodable format
        create_id: bool
            If True, creates ID for users automatically

        Example
        ----------
        .. code-block::

            from relevanceai import Client

            client = Client()

            documents = [
                {
                    "_id": "321",
                    "value": 10
                },
                {
                    "_id": "4243",
                    "value": 100
                }
            ]

            dataset_id = "sample_dataset_id"
            ds = client.Dataset(dataset_id)
            ds.upsert_documents(documents)

        """
        results = self._update_documents(
            self.dataset_id,
            documents=documents,
            bulk_fn=bulk_fn,
            max_workers=max_workers,
            retry_chunk_mult=retry_chunk_mult,
            show_progress_bar=show_progress_bar,
            chunksize=chunksize,
            use_json_encoder=use_json_encoder,
            create_id=create_id,
        )
        return self._process_insert_results(results, return_json=return_json)

    @track
    def apply(
        self,
        func: Callable,
        retrieve_chunksize: int = 100,
        filters: Optional[list] = None,
        select_fields: Optional[list] = None,
        show_progress_bar: bool = True,
        use_json_encoder: bool = True,
        axis: int = 0,
        log_to_file: bool = True,
        **apply_args,
    ):
        """
        Apply a function along an axis of the DataFrame.

        Objects passed to the function are Series objects whose index is either the DataFrame’s index (axis=0) or the DataFrame’s columns (axis=1). By default (result_type=None), the final return type is inferred from the return type of the applied function. Otherwise, it depends on the result_type argument.

        Parameters
        --------------
        func: function
            Function to apply to each document
        retrieve_chunksize: int
            The number of documents that are received from the original collection with each loop iteration.
        max_workers: int
            The number of processors you want to parallelize with
        max_error: int
            How many failed uploads before the function breaks
        json_encoder : bool
            Whether to automatically convert documents to json encodable format
        axis: int
            Axis along which the function is applied.
            - 9 or 'index': apply function to each column
            - 1 or 'columns': apply function to each row

        Example
        ---------
        .. code-block::

            from relevanceai import Client
            from relevanceai.package_utils.datasets import mock_documents

            client = Client()

            ds = client.Dataset("sample_dataset_id")
            ds.upsert_documents(mock_documents(100))

            def update_doc(doc):
                doc["value"] = 2
                return doc

            df.apply(update_doc)

            def update_doc_wargs(doc, value1, value2):
                doc["value"] += value1
                doc["value"] *= value2
                return doc

            df.apply(func=update_doc, value1=3, value2=2)

        """
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        if axis == 1:
            raise ValueError("We do not support column-wise operations!")

        def bulk_fn(documents):
            new_documents = []
            for d in documents:
                new_d = func(d, **apply_args)
                new_documents.append(new_d)
            return documents

        results = self.pull_update_push_async(
            self.dataset_id,
            bulk_fn,
            retrieve_chunk_size=retrieve_chunksize,
            filters=filters,
            select_fields=select_fields,
            show_progress_bar=show_progress_bar,
            use_json_encoder=use_json_encoder,
            log_to_file=log_to_file,
        )
        if results is None:
            print("✅ Successfully ran!")
            return
        for k, v in results.items():
            if v != []:
                print("️❗❗Errors detected when running apply.")
                return results
        print("✅ Successfully ran!")

    @track
    def bulk_apply(
        self,
        bulk_func: Callable,
        retrieve_chunksize: int = 100,
        filters: Optional[list] = None,
        select_fields: Optional[list] = None,
        show_progress_bar: bool = True,
        use_json_encoder: bool = True,
        log_to_file: bool = True,
    ):
        """
        Apply a bulk function along an axis of the DataFrame.

        Parameters
        ------------
        bulk_func: function
            Function to apply to a bunch of documents at a time
        retrieve_chunksize: int
            The number of documents that are received from the original collection with each loop iteration.
        max_workers: int
            The number of processors you want to parallelize with
        max_error: int
            How many failed uploads before the function breaks
        json_encoder : bool
            Whether to automatically convert documents to json encodable format
        axis: int
            Axis along which the function is applied.
            - 9 or 'index': apply function to each column
            - 1 or 'columns': apply function to each row

        Example
        ---------
        .. code-block::

            from relevanceai import Client

            client = Client()

            df = client.Dataset("sample_dataset_id")

            def update_documents(documents):
                for d in documents:
                    d["value"] = 10
                return documents

            df.apply(update_documents)
        """
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        return self.pull_update_push_async(
            self.dataset_id,
            bulk_func,
            retrieve_chunk_size=retrieve_chunksize,
            filters=filters,
            select_fields=select_fields,
            show_progress_bar=show_progress_bar,
            use_json_encoder=use_json_encoder,
            log_to_file=log_to_file,
        )

    @track
    def cat(self, vector_name: Union[str, None] = None, fields: Optional[List] = None):
        """
        Concatenates numerical fields along an axis and reuploads this vector for other operations

        Parameters
        ----------
        vector_name: str, default None
            name of the new concatenated vector field
        fields: List
            fields alone which the new vector will concatenate

        Example
        -----------------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            fields = [
                "numeric_field1",
                "numeric_field2",
                "numeric_field3"
            ]

            df.concat(fields)

            concat_vector_field_name = "concat_vector_"
            df.concat(vector_name=concat_vector_field_name, fields=fields)
        """
        fields = [] if fields is None else fields

        if vector_name is None:
            vector_name = "_".join(fields) + "_cat_vector_"

        def cat_fields(documents, field_name):
            cat_vector_documents = [
                {"_id": sample["_id"], field_name: [sample[field] for field in fields]}
                for sample in documents
            ]
            return cat_vector_documents

        self.pull_update_push(
            self.dataset_id, cat_fields, updating_args={"field_name": vector_name}
        )

    concat = cat

    def _label_cluster(self, label: Union[int, str]):
        if isinstance(label, (int, float)):
            return "cluster-" + str(label)
        return str(label)

    def _label_clusters(self, labels):
        return [self._label_cluster(x) for x in labels]

    def set_cluster_labels(self, vector_fields, alias, labels):
        def add_cluster_labels(documents):
            documents = self.get_all_documents(self.dataset_id)
            documents = list(filter(DocUtils.list_doc_fields, documents))
            set_cluster_field = (
                "_cluster_" + ".".join(vector_fields).lower() + "." + alias
            )
            self.set_field_across_documents(
                set_cluster_field,
                self._label_clusters(list(labels)),
                documents,
            )
            return documents

        self.pull_update_push(self.dataset_id, add_cluster_labels)

    @track
    def create(self, schema: Optional[dict] = None) -> Dict:
        """
        A dataset can store documents to be searched, retrieved, filtered and aggregated (similar to Collections in MongoDB, Tables in SQL, Indexes in ElasticSearch).
        A powerful and core feature of VecDB is that you can store both your metadata and vectors in the same document. When specifying the schema of a dataset and inserting your own vector use the suffix (ends with) "_vector_" for the field name, and specify the length of the vector in dataset_schema. \n

        For example:

        .. code-block::
            {
                "product_media_vector_": 1024,
                "product_text_description_vector_" : 128
            }

        These are the field types supported in our datasets: ["text", "numeric", "date", "dict", "chunks", "vector", "chunkvector"]. \n

        For example:

        .. code-block::

            {
                "product_text_description" : "text",
                "price" : "numeric",
                "created_date" : "date",
                "product_texts_chunk_": "chunks",
                "product_text_chunkvector_" : 1024
            }

        You don't have to specify the schema of every single field when creating a dataset, as VecDB will automatically detect the appropriate data type for each field (vectors will be automatically identified by its "_vector_" suffix). Infact you also don't always have to use this endpoint to create a dataset as /datasets/bulk_insert will infer and create the dataset and schema as you insert new documents. \n

        Note:

            - A dataset name/id can only contain undercase letters, dash, underscore and numbers.
            - "_id" is reserved as the key and id of a document.
            - Once a schema is set for a dataset it cannot be altered. If it has to be altered, utlise the copy dataset endpoint.

        For more information about vectors check out the 'Vectorizing' section, services.search.vector or out blog at https://relevance.ai/blog. For more information about chunks and chunk vectors check out services.search.chunk.

        Parameters
        ----------
        schema : dict
            Schema for specifying the field that are vectors and its length

        Example
        ----------
        .. code-block::

            from relevanceai import Client
            client = Client()

            documents = [
                {
                    "_id": "321",
                    "value": 10
                },
                {
                    "_id": "4243",
                    "value": 100
                }
            ]

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.create()

            df.insert_documents(documents)
        """
        schema = {} if schema is None else schema

        return self.datasets.create(self.dataset_id, schema=schema)

    @track
    def delete(self):
        """
        Delete a dataset

        Example
        ---------
        .. code-block::

            from relevanceai import Client
            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            df.delete()

        """
        return self.datasets.delete(self.dataset_id)

    insert_df = insert_pandas_dataframe

    def _upload_media(
        self, presigned_url: str, media_content: bytes, verbose: bool = True
    ):
        if not isinstance(media_content, bytes):
            raise ValueError(
                f"media needs to be in a bytes format. Currently in {type(media_content)}"
            )
        response = requests.put(presigned_url, data=media_content)
        if response.status_code == 200:
            if verbose:
                print("media successfully uploaded.")

    def insert_media_url(self, media_url: str, verbose: bool = True):
        """
        Insert a single media URL
        """
        # media to download
        response = self.datasets.get_file_upload_urls(
            self.dataset_id, files=[media_url]
        )
        url = response["files"][0]["url"]
        self._upload_media(
            presigned_url=response["files"][0]["upload_url"],
            media_content=requests.get(media_url).content,
        )
        if verbose:
            print(f"media is hosted at {url}")
        return url

    def insert_media_urls(
        self,
        media_urls: List[str],
        verbose: bool = True,
        file_log: str = "insert_media_urls.log",
    ):
        """
        Insert a single media URL
        """
        # media to download
        response = self.datasets.get_file_upload_urls(self.dataset_id, files=media_urls)
        response_docs: dict = {"media_documents": [], "failed_medias": []}
        with FileLogger(file_log):
            for i, im in enumerate(tqdm(media_urls)):
                response_doc = {}
                response_doc["media_file"] = im
                response_doc["media_url"] = response["files"][i]["url"]
                try:
                    self._upload_media(
                        presigned_url=response["files"][i]["upload_url"],
                        media_content=requests.get(im).content,
                        verbose=verbose,
                    )
                    response_docs["media_documents"].append(response_doc)
                except Exception as e:
                    if verbose:
                        print(f"Failed to upload {im}.")
                    if verbose:
                        print(e)
                    response_docs["failed_medias"].append(response_doc)
        # Return the media URLs
        return response_docs

    def _open_local_media(self, fn: str) -> bytes:
        with open(fn, "rb") as fn_byte:
            f = fn_byte.read()
            b = bytes(f)
        return b

    def insert_local_media(self, media_fn: str, verbose: bool = True):
        """
        Insert local media

        Parameters
        -------------
        media_fn: str
            A local media to upload
        verbose: bool
            If True, prints a statement after uploading each media
        """
        # media to download
        response = self.datasets.get_file_upload_urls(self.dataset_id, files=[media_fn])
        url = response["files"][0]["url"]
        self._upload_media(
            presigned_url=response["files"][0]["upload_url"],
            media_content=self._open_local_media(media_fn),
            verbose=verbose,
        )
        if verbose:
            print(f"media is hosted at {url}.")
        return url

    def insert_local_medias(
        self,
        media_fns: List[str],
        verbose: bool = False,
        file_log="local_media_upload.log",
    ):
        """Insert a list of local medias.

        Parameters
        ------------
        media_fns: List[str]
            A list of local medias
        verbose: bool
            If True, this will print after each successful upload.
        file_log: str
            The log to write
        """
        response = self.datasets.get_file_upload_urls(self.dataset_id, files=media_fns)
        response_docs: dict = {"media_documents": [], "failed_medias": []}
        with FileLogger(file_log) as f:
            for i, media_fn in enumerate(tqdm(media_fns)):
                response_doc = {}
                response_doc["media_file"] = media_fn
                response_doc["media_url"] = response["files"][i]["url"]
                try:
                    self._upload_media(
                        presigned_url=response["files"][i]["upload_url"],
                        media_content=self._open_local_media(media_fn),
                        verbose=verbose,
                    )
                    response_docs["media_documents"].append(response_doc)
                except Exception as e:
                    print(f"failed to upload {media_fn}")
                    print(e)
                    response_docs["failed_medias"].append(response_doc)
        return response_docs

    def get_media_documents(
        self,
        media_fns: List[str],
        verbose: bool = False,
        file_log: str = "media_upload.log",
    ) -> dict:
        """
        Bulk insert medias. Returns a link to once it has been hosted

        Parameters
        --------------
        media_fns: List[str]
            List of medias to upload
        verbose: bool
            If True, prints statements after uploading
        file_log: str
            The file log to write
        """
        # Algorithm aims to insert local or hosted medias
        if "http" in media_fns[0]:
            return self.insert_media_urls(media_fns, verbose=verbose, file_log=file_log)
        else:
            return self.insert_local_medias(
                media_fns, verbose=verbose, file_log=file_log
            )

    def upsert_media(
        self,
        media_fns: List[str],
        verbose: bool = False,
        file_log: str = "media_upload.log",
        **kw,
    ):
        """
        Insert medias into a dataset.

        Parameters
        -------------

        media_fns: List[str]
            A list of medias to upsert
        verbose: bool
            If True, prints statements after uploading
        file_log: str
            The file log to write
        """
        documents = self.get_media_documents(
            media_fns=media_fns, verbose=verbose, file_log=file_log
        )
        return self.upsert_documents(documents["media_documents"], create_id=True, **kw)

    def delete_documents(self, document_ids: List[str]):
        """
        Delete documents in a dataset

        Parameters
        ------------
        document_ids: List[str]
            A list of document IDs to delete

        """
        return self.datasets.documents.bulk_delete(
            dataset_id=self.dataset_id, ids=document_ids
        )



import pytest

from typing import Dict, List

from relevanceai import Client

from tests.globals.constants import generate_dataset_id


@pytest.fixture(scope="function")
def pandas_dataset(
    test_client: Client,
    pandas_documents: List[Dict],
):
    test_dataset_id = generate_dataset_id()

    response = test_client._insert_documents(test_dataset_id, pandas_documents)

    yield response, len(pandas_documents)

    test_client.datasets.delete(test_dataset_id)



import pytest

from relevanceai.client import Client
from relevanceai.dataset import Dataset


@pytest.fixture(scope="function")
def test_dataset(
    test_client: Client,
    vector_dataset_id: str,
) -> Dataset:
    dataset = test_client.Dataset(vector_dataset_id)
    yield dataset
    dataset.delete()


@pytest.fixture(scope="function")
def test_nested_assorted_df(
    test_client: Client,
    assorted_nested_dataset: str,
) -> Dataset:
    dataset = test_client.Dataset(assorted_nested_dataset)
    return dataset


@pytest.fixture(scope="function")
def test_sample_obj_dataset_id_df(
    test_client: Client,
    obj_dataset_id: str,
) -> Dataset:
    dataset = test_client.Dataset(obj_dataset_id)
    return dataset


@pytest.fixture(scope="function")
def test_clustered_df(
    test_client: Client,
    clustered_dataset_id: str,
) -> Dataset:
    dataset = test_client.Dataset(clustered_dataset_id)
    return dataset



from relevanceai.client.helpers import Credentials
from relevanceai.constants import CONFIG
from relevanceai.utils.transport import Transport
from relevanceai.utils.logger import LoguruLogger


def str2bool(v: str):
    return v.lower() in ("yes", "true", "t", "1")


class _Base(Transport, LoguruLogger):
    """_Base class for all relevanceai client utilities"""

    def __init__(
        self,
        credentials: Credentials,
        **kwargs,
    ):
        self.project = credentials.project
        self.api_key = credentials.api_key
        self.firebase_uid = credentials.firebase_uid

        self.config = CONFIG
        # Initialize logger
        super().__init__(**kwargs)

    ### Configurations

    @property
    def base_url(self):
        return CONFIG.get_field("api.base_url", CONFIG.config)

    @base_url.setter
    def base_url(self, value: str):
        if value.endswith("/"):
            value = value[:-1]
        CONFIG.set_option("api.base_url", value)

    @property
    def mixpanel_write_key(self):
        return CONFIG.get_field("mixpanel.write_key", CONFIG.config)

    @property
    def base_ingest_url(self):
        return CONFIG.get_field("api.base_ingest_url", CONFIG.config)

    @base_ingest_url.setter
    def base_ingest_url(self, value: str):
        if value.endswith("/"):
            value = value[:-1]
        CONFIG.set_option("api.base_ingest_url", value)

    @property
    def region(self):
        if hasattr(self, "_region"):
            return self._region
        return CONFIG["api.region"]

    @region.setter
    def region(self, region_value):
        CONFIG["api.region"] = region_value
        self._region = region_value



import numpy as np

from doc_utils import DocUtils
from abc import abstractmethod, ABC
from typing import Union, List, Dict, Callable

from relevanceai.utils.integration_checks import (
    is_hdbscan_available,
    is_sklearn_available,
)


class ClusterBase(DocUtils, ABC):
    """
    A Cluster _Base for models to be inherited.
    The most basic class to inherit.
    Use this class if you have an in-memory fitting algorithm.

    If your clusters return centroids, you will want to inherit
    `CentroidClusterBase`.

    If your clusters can fit on batches, you will want to inherit
    `BatchClusterBase`.

    If you have both Batches and Centroids, you will want to inherit both.

    """

    def __call__(self, *args, **kwargs):
        return self.fit_predict(*args, **kwargs)

    @abstractmethod
    def fit_predict(self, vectors: list) -> List[Union[str, float, int]]:
        """Edit this method to implement a ClusterBase.

        Parameters
        -------------
        vectors: list
            The vectors that are going to be clustered

        Example
        ----------

        .. code-block::

            class KMeansModel(ClusterBase):
                def __init__(self, k=10, init="k-means++", n_init=10,
                    max_iter=300, tol=1e-4, verbose=0, random_state=None,
                        copy_x=True,algorithm="auto"):
                        self.init = init
                        self.n_init = n_init
                        self.max_iter = max_iter
                        self.tol = tol
                        self.verbose = verbose
                        self.random_state = random_state
                        self.copy_x = copy_x
                        self.algorithm = algorithm
                        self.n_clusters = k

            def _init_model(self):
                from sklearn.cluster import KMeans
                self.km = KMeans(
                    n_clusters=self.n_clusters,
                    init=self.init,
                    verbose=self.verbose,
                    max_iter=self.max_iter,
                    tol=self.tol,
                    random_state=self.random_state,
                    copy_x=self.copy_x,
                    algorithm=self.algorithm,
                )
                return

            def fit_predict(self, vectors: Union[np.ndarray, List]):
                if not hasattr(self, "km"):
                    self._init_model()
                self.km.fit(vectors)
                cluster_labels = self.km.labels_.tolist()
                # cluster_centroids = km.cluster_centers_
                return cluster_labels

        """
        raise NotImplementedError

    def _concat_vectors_from_list(self, list_of_vectors: list):
        return [np.concatenate(x) for x in list_of_vectors]

    def _get_vectors_from_documents(self, vector_fields, docs):
        if len(vector_fields) == 1:
            # filtering out entries not containing the specified vector
            docs = list(filter(DocUtils.list_doc_fields, docs))
            vectors = self.get_field_across_documents(
                vector_fields[0], docs, missing_treatment="skip"
            )
        else:
            # In multifield clusering, we get all the vectors in each document
            # (skip if they are missing any of the vectors)
            # Then run clustering on the result
            docs = list(self.filter_docs_for_fields(vector_fields, docs))
            all_vectors = self.get_fields_across_documents(
                vector_fields, docs, missing_treatment="skip_if_any_missing"
            )
            # Store the vector field lengths to de-concatenate them later
            self._vector_field_length: dict = {}
            prev_vf = 0
            for i, vf in enumerate(self.vector_fields):
                self._vector_field_length[vf] = {}
                self._vector_field_length[vf]["start"] = prev_vf
                end_vf = prev_vf + len(all_vectors[0][i])
                self._vector_field_length[vf]["end"] = end_vf
                # Update the ending
                prev_vf = end_vf

            # Store the vector lengths
            vectors = self._concat_vectors_from_list(all_vectors)
        return vectors

    def __getdoc__(self, documents):
        """What you want in each doc"""
        raise NotImplementedError

    def _bulk_get_doc(self, documents):
        raise NotImplementedError

    def fit_documents(
        self,
        vector_fields: list,
        documents: List[dict],
        alias: str = "default",
        cluster_field: str = "_cluster_",
        return_only_clusters: bool = True,
        inplace: bool = True,
    ):
        """
        Train clustering algorithm on documents and then store the labels
        inside the documents.

        Parameters
        -----------
        vector_field: list
            The vector field of the documents
        documents: list
            List of documents to run clustering on
        alias: str
            What the clusters can be called
        cluster_field: str
            What the cluster fields should be called
        return_only_clusters: bool
            If True, return only clusters, otherwise returns the original document
        inplace: bool
            If True, the documents are edited inplace otherwise, a copy is made first
        kwargs: dict
            Any other keyword argument will go directly into the clustering algorithm

        """
        self.vector_fields = vector_fields

        vectors = self._get_vectors_from_documents(vector_fields, documents)

        cluster_labels = self.fit_predict(vectors)

        # Label the clusters
        cluster_labels = self._label_clusters(cluster_labels)

        if isinstance(vector_fields, list):
            set_cluster_field = f"{cluster_field}.{'.'.join(vector_fields)}.{alias}"
        elif isinstance(vector_fields, str):
            set_cluster_field = f"{cluster_field}.{vector_fields}.{alias}"

        if inplace:
            self.set_field_across_documents(
                set_cluster_field,
                cluster_labels,
                documents,
            )
            if return_only_clusters:
                return [
                    {"_id": d.get("_id"), cluster_field: d.get(cluster_field)}
                    for d in documents
                ]
            return documents

        new_documents = documents.copy()

        self.set_field_across_documents(
            set_cluster_field, cluster_labels, new_documents
        )

        if return_only_clusters:
            return [
                {"_id": d.get("_id"), cluster_field: d.get(cluster_field)}
                for d in documents
            ]
        return documents

    @property
    def metadata(self) -> dict:
        """If metadata is set - this willi be stored on RelevanceAI.
        This is useful when you are looking to compare the metadata of your clusters.
        """
        return {}

    def _label_cluster(self, label: Union[int, str]):
        if not isinstance(label, str):
            return "cluster-" + str(label)
        return label

    def _label_clusters(self, labels):
        return [self._label_cluster(x) for x in labels]


class AdvancedCentroidClusterBase(ClusterBase, ABC):
    """
    This centroid cluster base assumes that you want to specify
    quite advanced centroid documents.

    You may want to use this if you want to get more control over
    what is actually inserted as a centroid.
    """

    @abstractmethod
    def get_centroid_documents(self) -> List[Dict]:
        """Get the centroid documents."""
        pass


class CentroidBase(ABC):
    """
    Simple centroid base for clusters.
    """

    vector_fields: list
    _label_cluster: Callable

    @abstractmethod
    def get_centers(self) -> List[List[float]]:
        """Add how you need to get centers here. This should return a list of vectors.
        The SDK will then label each center `cluster-0`, `cluster-1`, `cluster-2`, etc... in order.
        If you need more fine-grained control, please see get_centroid_documents.
        """
        pass

    def get_centroid_documents(self) -> List:
        """
        Get the centroid documents to store. This enables you to use `list_closest_to_center()`
        and `list_furthest_from_center`.

        .. code-block::

            {
                "_id": "document-id-1",
                "centroid_vector_": [0.23, 0.24, 0.23]
            }

        If multiple vector fields returns this:
        Returns multiple

        .. code-block::

            {
                "_id": "document-id-1",
                "blue_vector_": [0.12, 0.312, 0.42],
                "red_vector_": [0.23, 0.41, 0.3]
            }

        """
        self.centers = self.get_centers()

        if not hasattr(self, "vector_fields") or len(self.vector_fields) == 1:
            if isinstance(self.centers, np.ndarray):
                self.centers = self.centers.tolist()
            centroid_vector_field_name = self.vector_fields[0]
            return [
                {
                    "_id": self._label_cluster(i),
                    centroid_vector_field_name: self.centers[i],
                }
                for i in range(len(self.centers))
            ]
        # For one or more vectors, separate out the vector fields
        # centroid documents are created using multiple vector fields
        centroid_docs = []
        for i, c in enumerate(self.centers):
            centroid_doc = {"_id": self._label_cluster(i)}
            for j, vf in enumerate(self.vector_fields):
                centroid_doc[vf] = self.centers[i][vf]
            centroid_docs.append(centroid_doc.copy())
        return centroid_docs


if is_sklearn_available():
    from sklearn.cluster import KMeans


class SklearnCentroidBase(CentroidBase, ClusterBase):
    def __init__(self, model):
        self.model: KMeans = model

    def get_centers(self):
        if hasattr(self.model, "cluster_centers_"):
            return self.model.cluster_centers_
        # Get the centers for each label
        centers = []
        labels = self.get_unique_labels()
        for l in sorted(np.unique(labels).tolist()):
            # self.model.jkjkj
            centers.append(self._X[self.preds == l].mean(axis=0).tolist())
        return centers

    def fit_predict(self, X):
        self._X = np.array(X)
        self.preds = self.model.fit_predict(X)
        return self.preds

    def get_unique_labels(self):
        if hasattr(self.model, "_labels"):
            labels = self.model._labels
        # Get labels from hdbscan
        elif hasattr(self.model, "labels_"):
            labels = self.model.labels_
        else:
            raise AttributeError(
                "SKLearn has changed labels API - will need to provide way to return cluster centers"
            )
        return sorted(np.unique(labels))

    def get_centroid_documents(self) -> List:
        """
        Get the centroid documents to store. This enables you to use `list_closest_to_center()`
        and `list_furthest_from_center`.

        .. code-block::

            {
                "_id": "document-id-1",
                "centroid_vector_": [0.23, 0.24, 0.23]
            }

        If multiple vector fields returns this:
        Returns multiple

        .. code-block::

            {
                "_id": "document-id-1",
                "blue_vector_": [0.12, 0.312, 0.42],
                "red_vector_": [0.23, 0.41, 0.3]
            }

        """
        self.centers = self.get_centers()

        if not hasattr(self, "vector_fields") or len(self.vector_fields) == 1:
            if isinstance(self.centers, np.ndarray):
                self.centers = self.centers.tolist()
            centroid_vector_field_name = self.vector_fields[0]
            return [
                {
                    "_id": self._label_cluster(self.model.labels_[i]),
                    centroid_vector_field_name: self.centers[i],
                }
                for i in range(len(self.centers))
            ]
        # For one or more vectors, separate out the vector fields
        # centroid documents are created using multiple vector fields
        centroid_docs = []
        for i, c in enumerate(self.centers):
            centroid_doc = {"_id": self._label_cluster(i)}
            for j, vf in enumerate(self.vector_fields):
                centroid_doc[vf] = self.centers[i][vf]
            centroid_docs.append(centroid_doc.copy())
        return centroid_docs


if is_hdbscan_available():
    import hdbscan


class HDBSCANClusterBase(SklearnCentroidBase):
    model: "hdbscan.HDBSCAN"

    def get_unique_labels(self):
        return sorted(np.unique(self.model.labels_))

    def get_centers(self):
        labels = self.get_unique_labels()
        centers = []
        for l in labels:
            centers.append(
                self.model._raw_data[self.model.labels_ == l].mean(axis=0).tolist()
            )
        return centers

    def get_centroid_documents(self) -> List:
        """
        Get the centroid documents to store. This enables you to use `list_closest_to_center()`
        and `list_furthest_from_center`.

        .. code-block::

            {
                "_id": "document-id-1",
                "centroid_vector_": [0.23, 0.24, 0.23]
            }

        If multiple vector fields returns this:
        Returns multiple

        .. code-block::

            {
                "_id": "document-id-1",
                "blue_vector_": [0.12, 0.312, 0.42],
                "red_vector_": [0.23, 0.41, 0.3]
            }

        """
        self.centers = self.get_centers()
        labels = self.get_unique_labels()

        if not hasattr(self, "vector_fields") or len(self.vector_fields) == 1:

            if isinstance(self.centers, np.ndarray):
                self.centers = self.centers.tolist()

            centroid_vector_field_name = self.vector_fields[0]
            return [
                {
                    "_id": self._label_cluster(labels[i]),
                    centroid_vector_field_name: self.centers[i],
                }
                for i in range(len(self.centers))
            ]
        # For one or more vectors, separate out the vector fields
        # centroid documents are created using multiple vector fields
        centroid_docs = []
        for i, c in enumerate(self.centers):
            centroid_doc = {"_id": self._label_cluster(labels[i])}
            for j, vf in enumerate(self.vector_fields):
                centroid_doc[vf] = self.centers[i][vf]
            centroid_docs.append(centroid_doc.copy())
        return centroid_docs


class CentroidClusterBase(ClusterBase, CentroidBase, ABC):
    """
    Inherit this class if you have a centroids-based clustering.
    The difference between this and `Clusterbase` is that you can also additionally
    specify how to get your centers in the
    `get_centers` base. This allows you to store your centers.
    """

    ...


class BatchClusterBase(ClusterBase, ABC):
    """
    Inherit this class if you have a batch-fitting algorithm that needs to be
    trained and then predicted separately.
    """

    @abstractmethod
    def partial_fit(self, vectors):
        """
        Partial fit the vectors.
        """
        pass

    @abstractmethod
    def predict(self):
        """
        Predict the vectors.
        """
        pass

    def fit_predict(self, X):
        pass



from typing import Optional
from relevanceai.client.helpers import Credentials

from relevanceai.utils.base import _Base


class AggregateClient(_Base):
    """Aggregate service"""

    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    def aggregate(
        self,
        dataset_id: str,
        metrics: Optional[list] = None,
        groupby: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        alias: str = "default",
        **kw
    ):
        """
        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:

        >>> {
        >>>        "groupby" : [
        >>>            {"name": <alias>, "field": <field in the collection>, "agg": "category"},
        >>>            {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
        >>>        ],
        >>>        "metrics" : [
        >>>            {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
        >>>            {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
        >>>        ]
        >>>    }
        >>>    For example, one can use the following aggregations to group score based on region and player name.
        >>>    {
        >>>        "groupby" : [
        >>>            {"name": "region", "field": "player_region", "agg": "category"},
        >>>            {"name": "player_name", "field": "name", "agg": "category"}
        >>>        ],
        >>>        "metrics" : [
        >>>            {"name": "average_score", "field": "final_score", "agg": "avg"},
        >>>            {"name": "max_score", "field": "final_score", "agg": "max"},
        >>>            {'name':'total_score','field':"final_score", 'agg':'sum'},
        >>>            {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
        >>>            {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
        >>>        ]
        >>>    }

        "groupby" is the fields you want to split the data into. These are the available groupby types:

            - category : groupby a field that is a category
            - numeric: groupby a field that is a numeric

        "metrics" is the fields and metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types:

            - "avg", "max", "min", "sum", "cardinality"

        The response returned has the following in descending order. \n

        If you want to return documents, specify a "group_size" parameter and a "select_fields" parameter if you want to limit the specific fields chosen. This looks as such:

        >>>    {
        >>>    'groupby':[
        >>>        {'name':'Manufacturer','field':'manufacturer','agg':'category',
        >>>        'group_size': 10, 'select_fields': ["name"]},
        >>>    ],
        >>>    'metrics':[
        >>>        {'name':'Price Average','field':'price','agg':'avg'},
        >>>    ],
        >>>    }
        >>>
        >>>    {"title": {"title": "books", "frequency": 200, "documents": [{...}, {...}]}, {"title": "books", "frequency": 100, "documents": [{...}, {...}]}}

        For array-aggregations, you can add "agg": "array" into the aggregation query.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias} vector
        """
        metrics = [] if metrics is None else metrics
        groupby = [] if groupby is None else groupby
        filters = [] if filters is None else filters

        return self.make_http_request(
            "/services/aggregate/aggregate",
            method="POST",
            parameters={
                "dataset_ids": [dataset_id],
                "aggregation_query": {"groupby": groupby, "metrics": metrics},
                "filters": filters,
                "page_size": page_size,
                "page": page,
                "asc": asc,
                "flatten": flatten,
                "alias": alias,
            },
            # base_url="https://gateway-api-aueast.relevance.ai/v1",
            **kw
        )



"""The Transport Class defines a transport as used by the Channel class to communicate with the network.
"""
import asyncio
import codecs
import time
import traceback

from json.decoder import JSONDecodeError
from typing import Optional

from urllib.parse import urlparse

import aiohttp
import requests

from requests import Request

from relevanceai.constants.config import Config
from relevanceai.utils.logger import AbstractLogger
from relevanceai.dashboard.dashboard_mappings import DASHBOARD_MAPPINGS
from relevanceai.constants.errors import APIError
from relevanceai.utils.json_encoder import JSONEncoderUtils

DO_NOT_REPEAT_STATUS_CODES = {404, 422}


class Transport(JSONEncoderUtils):
    """_Base class for all relevanceai objects"""

    project: str
    api_key: str
    config: Config
    logger: AbstractLogger

    @property
    def _dashboard_request_url(self):
        return self.config.get_option("dashboard.dashboard_request_url")

    @property
    def auth_header(self):
        return {"Authorization": self.project + ":" + self.api_key}

    @property
    def _search_dashboard_url(self):
        return (
            self.config["dashboard.base_dashboard_url"]
            + self.config["dashboard.search_dashboard_endpoint"]
        )

    @staticmethod
    def _is_search_in_path(url: str):
        if url is None:
            return False
        result = urlparse(url)
        split_path = result.path.split("/")
        return "search" in split_path and "services" in split_path

    @property
    def DASHBOARD_TYPES(self):
        return list(DASHBOARD_MAPPINGS.keys())

    def _log_to_dashboard(
        self,
        method: str,
        parameters: dict,
        endpoint: str,
        dashboard_type: str,
        verbose: bool = True,
    ):
        """Log search to dashboard"""
        # Needs to be a supported dashboard type
        if dashboard_type not in self.DASHBOARD_TYPES:
            return
        # Get the URL but not the version
        url = "/".join(self.config.get_option("api.base_url").split("/")[:-1]) + "/"
        # Split off the version separately
        version = self.config.get_option("api.base_url").split("/")[-1]
        # Parse the endpoint so it becomes 'endpoint/schema' instead of '/endpoint/schema'
        if endpoint.startswith("/"):
            endpoint = endpoint[1:]
        request_body = {
            dashboard_type: {
                "body": parameters,
                "url": url,
                "version": version,
                "endpoint": endpoint,
                "metadata": parameters,
                "query": parameters.get("query"),
            },
        }
        self.logger.debug(request_body)

        async def run_request():
            req = Request(
                method=method.upper(),
                url=self._dashboard_request_url,
                headers=self.auth_header,
                json=request_body,
                # params=parameters if method.upper() == "GET" else {},
            ).prepare()
            with requests.Session() as s:
                response = s.send(req)

        asyncio.ensure_future(run_request())

        if verbose:
            dashboard_url = (
                self.config["dashboard.base_dashboard_url"]
                + DASHBOARD_MAPPINGS[dashboard_type]
            )
            self.print_dashboard_url(dashboard_url)

    def _link_to_dataset_dashboard(self, dataset_id: str, suburl: str = None):
        """Link to a monitoring dashboard
        Suburl must be one of
        - "monitor"
        - "lookups"
        - "monitor/schema"
        """
        MESSAGE = "You can view your dashboard at: "
        if suburl is None:
            print(
                MESSAGE
                + f"https://cloud.relevance.ai/dataset/{dataset_id}/dashboard/monitor/"
            )
        else:
            print(
                MESSAGE
                + f"https://cloud.relevance.ai/dataset/{dataset_id}/dashboard/{suburl}"
            )

    def _log_search_to_dashboard(self, method: str, parameters: dict, endpoint: str):
        """Log search to dashboard"""
        self._log_to_dashboard(
            method=method,
            parameters=parameters,
            endpoint=endpoint,
            dashboard_type="multivector_search",
        )

    def print_dashboard_message(self, message: str):
        if self.config["dashboard.show_dashboard_link"]:
            print(message)

    def print_dashboard_url(self, dashboard_url):
        if self.config["dashboard.show_dashboard_link"]:
            print(f"You can now visit the dashboard at {dashboard_url}")

    def make_http_request(
        self,
        endpoint: str,
        method: str = "GET",
        parameters: Optional[dict] = None,
        base_url: str = None,
        output_format=None,
        raise_error: bool = True,
    ):
        """
        Make the HTTP request
        Parameters
        ----------
        endpoint: string
            The endpoint from the documentation to use
        method_type: string
            POST or GET request
        raise_error: bool
            If True, you will raise error. This is useful for endpoints that don't
            necessarily need to error.
        """
        parameters = {} if parameters is None else parameters
        self._last_used_endpoint = endpoint
        start_time = time.perf_counter()

        if base_url is None:
            # if Transport.is_search_in_path(base_url) and not hasattr(self, "output_format"):
            #     base_url = self.config.get_option("dashboard.base_dashboard_url")[1:-1]
            # else:
            base_url = self.config.get_option("api.base_url")

        if output_format is None:
            output_format = self.config.get_option("api.output_format")

        retries = int(self.config.get_option("retries.number_of_retries"))
        seconds_between_retries = int(
            self.config.get_option("retries.seconds_between_retries")
        )
        request_url = base_url + endpoint
        for _ in range(retries):

            self.logger.info("URL you are trying to access:" + request_url)
            try:
                if Transport._is_search_in_path(request_url):
                    self._log_search_to_dashboard(
                        method=method, parameters=parameters, endpoint=endpoint
                    )

                # TODO: Add other endpoints in here too
                if method.upper() in {"POST", "PUT"}:
                    req = Request(
                        method=method.upper(),
                        url=request_url,
                        headers=self.auth_header,
                        json=parameters if method.upper() == "POST" else {},
                    ).prepare()
                elif method.upper() == "GET":
                    # Get requests do not have JSONs - which will error out
                    # cloudfront
                    req = Request(
                        method=method.upper(),
                        url=request_url,
                        headers=self.auth_header,
                        params=parameters if method.upper() == "GET" else {},
                    ).prepare()

                with requests.Session() as s:
                    response = s.send(req)

                # Successful response
                if response.status_code == 200:
                    self._log_response_success(base_url, endpoint)
                    self._log_response_time(
                        base_url, endpoint, time.perf_counter() - start_time
                    )

                    if output_format == "json":
                        return response.json()
                    elif output_format == "content":
                        return response.content
                    elif output_format == "status_code":
                        return response.status_code
                    else:
                        return response

                # Cancel bad URLs
                # Logged status codes
                elif response.status_code in DO_NOT_REPEAT_STATUS_CODES:
                    self._log_response_fail(
                        base_url,
                        endpoint,
                        response.status_code,
                        response.content.decode(),
                    )
                    if raise_error:
                        raise APIError(response.content.decode())

                # Retry other errors
                else:
                    self._log_response_fail(
                        base_url,
                        endpoint,
                        response.status_code,
                        response.content.decode(),
                    )
                    continue

            except (ConnectionError) as error:
                # Print the error
                traceback.print_exc()
                self._log_connection_error(base_url, endpoint)
                time.sleep(seconds_between_retries)
                continue

            except JSONDecodeError as error:
                self._log_no_json(base_url, endpoint, response.status_code, response)
                return response

        return response

    async def make_async_http_request(
        self,
        endpoint: str,
        method: str = "GET",
        parameters: Optional[dict] = None,
        base_url: str = None,
        output_format=None,
        raise_error: bool = True,
    ):
        """
        Make an asynchronous HTTP request

        Parameters
        ----------
        endpoint: str
            The Relevance AI endpoint to access

        method_type: str
            Currently only support GET and POST requests

        raise_error: bool
            If True, error is raised rather than just logged.
        """
        parameters = {} if parameters is None else parameters
        self._last_used_endpoint = endpoint
        start_time = time.perf_counter()

        base_url = (
            self.config.get_option("api.base_url") if base_url is None else base_url
        )
        output_format = (
            self.config.get_option("api.output_format")
            if output_format is None
            else output_format
        )
        retries = int(self.config.get_option("retries.number_of_retries"))
        seconds_between_retries = int(
            self.config.get_option("retries.seconds_between_retries")
        )

        request_url = base_url + endpoint

        for _ in range(retries):
            self.logger.info(f"URL you are trying to access: {request_url}")
            try:
                if Transport._is_search_in_path(request_url):
                    self._log_search_to_dashboard(
                        method=method, parameters=parameters, endpoint=endpoint
                    )

                async with aiohttp.request(
                    method=method.upper(),
                    url=request_url,
                    headers=self.auth_header,
                    json=parameters if method.upper() == "POST" else {},
                    params=parameters if method.upper() == "GET" else {},
                ) as response:
                    if response.status == 200:
                        self._log_response_success(base_url, endpoint)
                        self._log_response_time(
                            base_url, endpoint, time.perf_counter() - start_time
                        )

                        if output_format.lower() == "json":
                            return await response.json()
                        elif output_format.lower() == "content":
                            decoded_content = codecs.decode(
                                await response.content.read()
                            )
                            return decoded_content
                        elif output_format.lower() == "status_code":
                            return response.status
                        else:
                            return response
                    elif response.status in DO_NOT_REPEAT_STATUS_CODES:
                        # Cancel bad URLs
                        # Logged status codes
                        decoded_content = codecs.decode(await response.content.read())
                        self._log_response_fail(
                            base_url, endpoint, response.status, decoded_content
                        )
                        if raise_error:
                            raise APIError(decoded_content)
                    else:
                        # Retry other errors
                        decoded_content = codecs.decode(await response.content.read())
                        self._log_response_fail(
                            base_url, endpoint, response.status, decoded_content
                        )
            except aiohttp.ClientError as error:
                traceback.print_exc()
                self._log_connection_error(base_url, endpoint)
                time.sleep(seconds_between_retries)
                continue
            except JSONDecodeError as error:
                self._log_no_json(base_url, endpoint, response.status, response)
                return response

        return response

    def _log_response_success(self, base_url, endpoint):
        self.logger.success(f"Response success! ({base_url + endpoint})")

    def _log_response_time(self, base_url, endpoint, time):
        self.logger.debug(f"Request ran in {time} seconds ({base_url + endpoint})")

    def _log_response_fail(self, base_url, endpoint, status_code, content):
        self.logger.error(
            f"Response failed ({base_url + endpoint}) (Status: {status_code} Response: {content})"
        )

    def _log_connection_error(self, base_url, endpoint):
        self.logger.error(f"Connection error but re-trying. ({base_url + endpoint})")

    def _log_no_json(self, base_url, endpoint, status_code, content):
        self.logger.error(
            f"No JSON Available ({base_url + endpoint}) (Status: {status_code} Response: {content})"
        )



"""All Dataset related functions
"""
from typing import Optional

from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base


class MonitorClient(_Base):
    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    def health(self, dataset_id: str):
        """
        Gives you a summary of the health of your vectors, e.g. how many documents with vectors are missing, how many documents with zero vectors

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        """
        # https://cloud.relevance.ai/dataset/demo-movies/dashboard/monitor/
        self._link_to_dataset_dashboard(dataset_id, "monitor/schema")
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/monitor/health", method="GET"
        )

    def stats(self, dataset_id: str):
        """
        All operations related to monitoring

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        """
        self._link_to_dataset_dashboard(dataset_id, "monitor/usage")
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/monitor/stats", method="GET"
        )

    def usage(
        self,
        dataset_id: str,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        log_ids: Optional[list] = None,
    ):
        """
        Aggregate the logs for a dataset. \n

        The response returned has the following fields:

        >>> [{'frequency': 958, 'insert_date': 1630159200000},...]

        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten	: bool
            Whether to flatten
        log_ids: list
            The log dataset IDs to aggregate with - one or more of logs, logs-write, logs-search, logs-task or js-logs
        """
        filters = [] if filters is None else filters
        log_ids = [] if log_ids is None else log_ids

        self._link_to_dataset_dashboard(dataset_id, "monitor/schema")
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/monitor/usage",
            method="POST",
            parameters={
                "filters": filters,
                "page_size": page_size,
                "page": page,
                "asc": asc,
                "filters": filters,
                "flatten": flatten,
                "log_ids": log_ids,
            },
        )



"""Relevance AI's base Client class - primarily used to login and access the Dataset class or ClusterOps class.  The recomended way to log in is using: .. code-block:: from relevanceai import Client client = Client() client.list_datasets()
If the user already knows their project and API key, they can
log in this way:

.. code-block::

    from relevanceai import Client
    project = ""
    api_key = ""
    client = Client(project=project, api_key=api_key, firebase_uid=firebase_uid)
    client.list_datasets()

If you need to change your token, simply run:

.. code-block::

    from relevanceai import Client
    client = Client(token="...")

"""
import getpass
import pandas as pd
import analytics

from base64 import b64decode as decode
from typing import Dict, List, Optional

from relevanceai._api import APIClient

from doc_utils.doc_utils import DocUtils

from relevanceai.client.helpers import *

from relevanceai.operations.cluster import ClusterOps
from relevanceai.constants.errors import APIError
from relevanceai.constants.messages import Messages
from relevanceai.dataset import Dataset

from relevanceai.utils.decorators.analytics import track, identify
from relevanceai.utils.decorators.version import beta, added


class Client(APIClient, DocUtils):
    def __init__(
        self,
        token: Optional[str] = None,
        authenticate: bool = True,
    ):
        """
        Initialize the client

        Parameters
        -------------

        token: str
            You can paste the token here if things need to be refreshed

        force_refresh: bool
            If True, it forces you to refresh your client
        """

        if token is None:
            token = auth()

        self.token = token
        self.credentials = process_token(token)
        # Eventually the following should be accessed directly from
        # self.credentials, but keep for now.
        (
            self.project,
            self.api_key,
            self.region,
            self.firebase_uid,
        ) = self.credentials.split_token()

        self.base_url = region_to_url(self.region)
        self.base_ingest_url = region_to_url(self.region)

        try:
            self._set_mixpanel_write_key()
        except Exception as e:
            pass

        self._identify()

        super().__init__(self.credentials)

        # used to debug
        if authenticate:
            if self.check_auth():
                print(Messages.WELCOME_MESSAGE.format(self.project))
            else:
                raise APIError(Messages.FAIL_MESSAGE)

        # Add non breaking changes to support old ways of inserting documents and csv
        self.insert_documents = Dataset(
            credentials=self.credentials,
            dataset_id="",
        )._insert_documents
        self.insert_csv = Dataset(
            credentials=self.credentials,
            dataset_id="",
        )._insert_csv

    @identify
    def _identify(self):
        return

    def _set_mixpanel_write_key(self):
        analytics.write_key = decode(self.mixpanel_write_key).decode("utf-8")

    def _get_token(self):
        # TODO: either use cache or keyring package
        token = getpass.getpass(f"Activation token:")
        return token

    @property
    def auth_header(self):
        return {"Authorization": self.project + ":" + self.api_key}

    def make_search_suggestion(self):
        return self.services.search.make_suggestion()

    def check_auth(self):
        print(f"Connecting to {self.region}...")
        return self.list_datasets()

    @track
    def create_dataset(self, dataset_id: str, schema: Optional[Dict] = None):
        """
        A dataset can store documents to be searched, retrieved, filtered and aggregated (similar to Collections in MongoDB, Tables in SQL, Indexes in ElasticSearch).
        A powerful and core feature of VecDB is that you can store both your metadata and vectors in the same document. When specifying the schema of a dataset and inserting your own vector use the suffix (ends with) "_vector_" for the field name, and specify the length of the vector in dataset_schema. \n

        For example:

        .. code-block::
            {
                "product_image_vector_": 1024,
                "product_text_description_vector_" : 128
            }

        These are the field types supported in our datasets: ["text", "numeric", "date", "dict", "chunks", "vector", "chunkvector"]. \n

        For example:

        .. code-block::

            {
                "product_text_description" : "text",
                "price" : "numeric",
                "created_date" : "date",
                "product_texts_chunk_": "chunks",
                "product_text_chunkvector_" : 1024
            }

        You don't have to specify the schema of every single field when creating a dataset, as VecDB will automatically detect the appropriate data type for each field (vectors will be automatically identified by its "_vector_" suffix). Infact you also don't always have to use this endpoint to create a dataset as /datasets/bulk_insert will infer and create the dataset and schema as you insert new documents. \n

        Note:

            - A dataset name/id can only contain undercase letters, dash, underscore and numbers.
            - "_id" is reserved as the key and id of a document.
            - Once a schema is set for a dataset it cannot be altered. If it has to be altered, utlise the copy dataset endpoint.

        For more information about vectors check out the 'Vectorizing' section, services.search.vector or out blog at https://relevance.ai/blog. For more information about chunks and chunk vectors check out services.search.chunk.

        Parameters
        ----------
        dataset_id: str
            The unique name of your dataset
        schema : dict
            Schema for specifying the field that are vectors and its length

        Example
        ----------
        .. code-block::

            from relevanceai import Client
            client = Client()
            client.create_dataset("sample_dataset_id")

        """
        schema = {} if schema is None else schema
        return self.datasets.create(dataset_id, schema=schema)

    @track
    def list_datasets(self):
        """List Datasets

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.list_datasets()

        """
        self.print_dashboard_message(
            "You can view all your datasets at https://cloud.relevance.ai/datasets/"
        )
        return self.datasets.list()

    @track
    def delete_dataset(self, dataset_id):
        """
        Delete a dataset

        Parameters
        ------------
        dataset_id: str
            The ID of a dataset

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.delete_dataset("sample_dataset_id")

        """
        return self.datasets.delete(dataset_id)

    @track
    def Dataset(
        self,
        dataset_id: str,
        fields: Optional[List[str]] = None,
        image_fields: Optional[List[str]] = None,
        audio_fields: Optional[List[str]] = None,
        highlight_fields: Optional[Dict[str, List]] = None,
        text_fields: Optional[List[str]] = None,
    ):
        fields = [] if fields is None else fields
        image_fields = [] if image_fields is None else image_fields
        audio_fields = [] if audio_fields is None else audio_fields
        highlight_fields = {} if highlight_fields is None else highlight_fields
        text_fields = [] if text_fields is None else text_fields
        return Dataset(
            credentials=self.credentials,
            dataset_id=dataset_id,
            fields=fields,
            image_fields=image_fields,
            audio_fields=audio_fields,
            highlight_fields=highlight_fields,
            text_fields=text_fields,
        )

    ### Clustering

    @track
    def ClusterOps(
        self,
        model,
        **kwargs,
    ):
        return ClusterOps(
            credentials=self.credentials,
            model=model,
            **kwargs,
        )

    def _set_logger_to_verbose(self):
        # Use this for debugging
        self.config["logging.logging_level"] = "INFO"

    @track
    def send_dataset(
        self,
        dataset_id: str,
        receiver_project: str,
        receiver_api_key: str,
    ):
        """
        Send an individual a dataset. For this, you must know their API key.


        Parameters
        -----------

        dataset_id: str
            The name of the dataset
        receiver_project: str
            The project name that will receive the dataset
        receiver_api_key: str
            The project API key that will receive the dataset


        Example
        --------

        .. code-block::

            client = Client()
            client.send_dataset(
                dataset_id="research",
                receiver_project="...",
                receiver_api_key="..."
            )

        """
        return self.admin.send_dataset(
            dataset_id=dataset_id,
            receiver_project=receiver_project,
            receiver_api_key=receiver_api_key,
        )

    @track
    def receive_dataset(
        self,
        dataset_id: str,
        sender_project: str,
        sender_api_key: str,
    ):
        """
        Recieve an individual a dataset.

        Example
        --------
        >>> client = Client()
        >>> client.admin.receive_dataset(
            dataset_id="research",
            sender_project="...",
            sender_api_key="..."
        )

        Parameters
        -----------

        dataset_id: str
            The name of the dataset
        sender_project: str
            The project name that will send the dataset
        sender_api_key: str
            The project API key that will send the dataset

        """
        return self.admin.receive_dataset(
            dataset_id=dataset_id,
            sender_project=sender_project,
            sender_api_key=sender_api_key,
        )

    @track
    def clone_dataset(
        self,
        source_dataset_id: str,
        new_dataset_id: Optional[str] = None,
        source_project: Optional[str] = None,
        source_api_key: Optional[str] = None,
        project: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        # To do this should be cloning a dataset in a project
        """
        Clone a dataset from another user's projects into your project.

        Parameters
        ----------
        dataset_id:
            The dataset to copy
        source_dataset_id:
            The original dataset
        source_project:
            The original project to copy from
        source_api_key:
            The original API key of the project
        project:
            The original project
        api_key:
            The original API key

        Example
        -----------

        .. code-block::

            client = Client()
            client.send_dataset(
                dataset_id="research",
                source_project="...",
                source_api_key="..."
            )

        """
        if source_api_key is None:
            source_api_key = self.api_key
            source_project = self.project

        if new_dataset_id is None:
            new_dataset_id = source_dataset_id

        return self.admin.copy_foreign_dataset(
            dataset_id=new_dataset_id,
            source_dataset_id=source_dataset_id,
            source_project=source_project,
            source_api_key=source_api_key,
            project=project,
            api_key=api_key,
        )

    @property
    def references(self):
        from relevanceai.__init__ import __version__

        REFERENCE_URL = f"https://relevanceai.readthedocs.io/en/{__version__}/"
        MESSAGE = f"You can find your references here {REFERENCE_URL}."
        print(MESSAGE)

    docs = references

    def search_app(self, dataset_id: Optional[str] = None):
        if dataset_id is not None:
            self.print_search_dashboard_url(dataset_id)
        elif hasattr(self, "_dataset_id"):
            self.print_search_dashboard_url(self._dataset_id)
        elif hasattr(self, "dataset_id"):
            self.print_search_dashboard_url(self.dataset_id)
        else:
            print("You can build your search app at https://cloud.relevance.ai")

    @added(version="1.1.3")
    @track
    def search_datasets(self, query: str):
        """
        Search through your datasets.
        """
        return [x for x in self.list_datasets()["datasets"] if query in x]

    @added(version="2.1.3")
    @beta
    def list_cluster_reports(self):
        """

        List all cluster reports.

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.list_cluster_reports()

        """
        return pd.DataFrame(self.reports.clusters.list()["results"])

    @added(version="2.1.3")
    @beta
    @track
    def delete_cluster_report(self, cluster_report_id: str):
        """

        Delete Cluster Report

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.delete_cluster_report("cluster_id_goes_here")

        """
        return self.reports.clusters.delete(cluster_report_id)

    @added(version="2.1.3")
    @beta
    @track
    def store_cluster_report(self, report_name: str, report: dict):
        """

        Store the cluster data.

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.store_cluster_report("sample", {"value": 3})

        """
        response: dict = self.reports.clusters.create(
            name=report_name, report=self.json_encoder(report)
        )
        print(
            f"You can now access your report at https://cloud.relevance.ai/report/cluster/{self.region}/{response['_id']}"
        )
        return response

    def disable_analytics_tracking(self):
        """Disable analytics tracking if you would prefer not to send usage
        data to improve the product. Analytics allows us to improve your experience
        by examining the most popular flows, dedicating more resources to popular
        product features and improve user experience.
        """
        self.config["mixpanel.is_tracking_enabled"] = False



from relevanceai.utils.decorators.analytics import track
from relevanceai.dataset.read import Read


class DictExport(Read):
    @track
    def to_dict(self, orient: str = "records"):
        """
        Returns the raw list of dicts from Relevance AI

        Parameters
        ----------
        None

        Returns
        -------
        list of documents in dictionary format

        Example
        -------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            dict = df.to_dict(orient="records")
        """
        if orient == "records":
            return self.get_all_documents()
        else:
            raise NotImplementedError



import os
from datetime import datetime
from setuptools import find_packages, setup


def read(rel_path):
    """Read lines from given file"""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    """Read __version__ from given file"""
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError(f"Unable to find a valid __version__ string in {rel_path}.")


requirements = [
    "tqdm>=4.49.0",
    "pandas>=1.0.0",
    "loguru>=0.5.3",
    "document-utils>=1.7.1",
    "requests>=2.0.0",
    "numpy>=1.19.0",
    "joblib>=1.0.0",
    "scikit-learn>=0.20.0",  # last version of support to Python3.4
    "typing-extensions~=4.0.1",
    "analytics-python~=1.4.0",
    "aiohttp>=3.8.1",
    "appdirs>=1.4.4",
    "orjson>=3.6.7",
]

excel_requirements = requirements + ["openpyxl>=3.0.9", "fsspec>=2021.10.1"]

vis_requirements = requirements + [
    "matplotlib>=3.5.1",
    "plotly>=5.5.0",
    "typeguard",
    "dash",
    "pillow",
    "opencv-python",
    "scikit-image",
    "dash_bootstrap_components",
]

umap = ["umap-learn>=0.5.2"]
# ivis_cpu = ["ivis[cpu]>=2.0.6"]
# ivis_gpu = ["ivis[gpu]>=2.0.6"]
kmedoids = ["scikit-learn-extra>=0.2.0"]
hdbscan = ["hdbscan>=0.8.27"]

# vis_extras = umap + ivis_cpu + ivis_gpu + kmedoids + hdbscan

test_requirements = (
    [
        "pytest",
        "pytest-dotenv",
        "pytest-xdist",
        "pytest-cov",
        "pytest-mock",
        "mypy",
        "types-requests",
        "pytest-sugar",
        "pytest-rerunfailures",
    ]
    + excel_requirements
    + vis_requirements
    + requirements
    # + vis_extras
)

doc_requirements = [
    "sphinx-rtd-theme>=0.5.0",
    "sphinx-autoapi==1.8.4",
    "sphinx-autodoc-typehints==1.12.0",
]

dev_requirements = (
    [
        "autopep8",
        "pylint",
        "jupyter",
    ]
    + test_requirements
    + doc_requirements
)


dev_vis_requirements = (
    ["autopep8", "pylint", "jupyter"]
    + test_requirements
    + vis_requirements
    + doc_requirements
    # + vis_extras
)

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

name = "RelevanceAI"
version = get_version("relevanceai/__init__.py")

if os.getenv("_IS_DEV"):
    name = "RelevanceAI-dev"
    version = (
        version
        + "."
        + datetime.now().__str__().replace("-", ".").replace(" ", ".").replace(":", ".")
    )

setup(
    name=name,
    version=version,
    url="https://relevance.ai/",
    author="Relevance AI",
    author_email="dev@relevance.ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    setup_requires=["wheel"],
    install_requires=requirements,
    package_data={
        "": [
            "*.ini",
        ]
    },
    extras_require={
        "docs": doc_requirements,
        "dev": dev_requirements,
        "dev-vis": dev_vis_requirements,
        "dev-viz": dev_vis_requirements,
        "excel": excel_requirements,
        "vis": vis_requirements,
        "viz": vis_requirements,
        # "vis-all": vis_requirements + vis_extras,
        "tests": test_requirements,
        "notebook": ["jsonshower"] + vis_requirements,
        "umap": umap,
        # "ivis-cpu": ivis_cpu,
        # "ivis-gpu": ivis_gpu,
        "kmedoids": kmedoids,
        "hdbscan": hdbscan,
    },
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Manufacturing",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Multimedia :: Sound/Audio :: Conversion",
        "Topic :: Multimedia :: Video :: Conversion",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



from typing import Dict, Any
from typing_extensions import Literal

from joblib.memory import Memory

dict = Dict[str, Any]

DIM_REDUCTION = Literal["pca", "tsne", "umap", "ivis"]
DIM_REDUCTION_DEFAULT_ARGS: Dict[Any, Any] = {
    "pca": {"svd_solver": "auto", "random_state": 42},
    "tsne": {
        "init": "pca",
        "n_iter": 500,
        "learning_rate": 100,
        "perplexity": 30,
        "random_state": 42,
    },
    "umap": {
        "n_neighbors": 10,
        "min_dist": 0.1,
        "random_state": 42,
        "transform_seed": 42,
    },
    "ivis": {"k": 15, "model": "maaten", "n_epochs_without_progress": 2},
}


GROUPBY_MAPPING = {"text": "category", "numeric": "numeric"}


CLUSTER = Literal["kmeans", "kmedoids", "hdbscan"]  # "kmodes", "kprototypes", None]
CLUSTER_DEFAULT_ARGS: Dict[str, Dict[str, Any]] = {
    "kmeans": {
        "k": 10,
        "init": "k-means++",
        "verbose": 0,
    },
    "kmedoids": {
        "metric": "euclidean",
        "init": "k-medoids++",
        "random_state": 42,
        "method": "pam",
    },
    "hdbscan": {
        "algorithm": "best",
        "alpha": 1.0,
        "approx_min_span_tree": True,
        "gen_min_span_tree": False,
        "leaf_size": 40,
        "memory": Memory(cachedir=None),
        "metric": "euclidean",
        "min_samples": None,
        "p": None,
    },
}

NEAREST_NEIGHBOURS = Literal["cosine", "l2"]
CENTROID_DISTANCES = Literal["cosine", "l2"]


SILHOUETTE_INFO = """
Good clusters have clusters which are highly seperated and elements within which are highly cohesive. <br/>
<b>Silohuette Score</b> is a metric from <b>-1 to 1</b> that calculates the average cohesion and seperation of each element, with <b>1</b> being clustered perfectly, <b>0</b> being indifferent and <b>-1</b> being clustered the wrong way"""

RAND_INFO = """Good clusters have elements, which, when paired, belong to the same cluster label and same ground truth label. <br/>
<b>Rand Index</b> is a metric from <b>0 to 1</b> that represents the percentage of element pairs that have a matching cluster and ground truth labels with <b>1</b> matching perfect and <b>0</b> matching randomly. <br/> <i>Note: This measure is adjusted for randomness so does not equal the exact numerical percentage.</i>"""

HOMOGENEITY_INFO = """Good clusters only have elements from the same ground truth within the same cluster<br/>
<b>Homogeneity</b> is a metric from <b>0 to 1</b> that represents whether clusters contain only elements in the same ground truth with <b>1</b> being perfect and <b>0</b> being absolutely incorrect."""

COMPLETENESS_INFO = """Good clusters have all elements from the same ground truth within the same cluster <br/>
<b>Completeness</b> is a metric from <b>0 to 1</b> that represents whether clusters contain all elements in the same ground truth with <b>1</b> being perfect and <b>0</b> being absolutely incorrect."""

METRIC_DESCRIPTION = {
    "Silhouette Score": SILHOUETTE_INFO,
    "Rand Score": RAND_INFO,
    "Homogeneity": HOMOGENEITY_INFO,
    "Completeness": COMPLETENESS_INFO,
}



from functools import singledispatch
from functools import update_wrapper


class singledispatchmethod:
    """Single-dispatch generic method descriptor.
    Supports wrapping existing descriptors and handles non-descriptor
    callables as instance methods.
    """

    def __init__(self, func):
        if not callable(func) and not hasattr(func, "__get__"):
            raise TypeError(f"{func!r} is not callable or a descriptor")

        self.dispatcher = singledispatch(func)
        self.func = func

    def register(self, cls, method=None):
        """generic_method.register(cls, func) -> func
        Registers a new implementation for the given *cls* on a *generic_method*.
        """
        return self.dispatcher.register(cls, func=method)

    def __get__(self, obj, cls=None):
        def _method(*args, **kwargs):
            method = self.dispatcher.dispatch(args[0].__class__)
            return method.__get__(obj, cls)(*args, **kwargs)

        _method.__isabstractmethod__ = self.__isabstractmethod__
        _method.register = self.register
        update_wrapper(_method, self.func)
        return _method

    @property
    def __isabstractmethod__(self):
        return getattr(self.func, "__isabstractmethod__", False)



from typing import Union, List, Any

from doc_utils import DocUtils

from relevanceai.client.helpers import Credentials
from relevanceai.operations.dr.dim_reduction import PCA
from relevanceai.operations.dr.dim_reduction import TSNE
from relevanceai.operations.dr.dim_reduction import Ivis
from relevanceai.operations.dr.dim_reduction import UMAP
from relevanceai._api import APIClient


class ReduceDimensionsOps(APIClient, DocUtils):
    def __init__(
        self,
        credentials: Credentials,
        n_components: int,
        model: Union[PCA, TSNE, Ivis, PCA, str, Any],
        dr_field: str = "_dr_",
        verbose: bool = True,
    ):
        if isinstance(model, str):
            algorithm = model.upper()
            if algorithm == "PCA":

                model = PCA()
            elif algorithm == "TSNE":

                model = TSNE()
            elif algorithm == "UMAP":

                model = UMAP()
            elif algorithm == "IVIS":

                model = Ivis()
            else:
                raise ValueError()

        self.model = model
        self.dr_field = dr_field
        self.verbose = verbose
        self.n_components = n_components

        super().__init__(credentials)

    def fit(
        self,
        dataset_id: str,
        vector_fields: str,
        alias: str,
    ):
        """
        Reduce Dimensions

        Parameters
        --------------

        fields: list
            The list of fields to run dimensionality reduction on. Currently
            only supports 1 field.
        documents: list
            The list of documents to run dimensionality reduction on
        inplace: bool
            If True, replaces the original documents, otherwise it returns
            a new set of documents with only the dr vectors in it and the _id
        """
        documents = self._get_all_documents(
            dataset_id, select_fields=vector_fields, include_vector=True
        )

        dr_documents = self.model.fit_transform_documents(
            vector_field=vector_fields[0],
            documents=documents,
            alias=alias,
            dims=self.n_components,
        )

        return self.update_documents(dataset_id, dr_documents)



from typing import Dict, List, Any, Union

from sklearn.preprocessing import (
    MinMaxScaler,
    MaxAbsScaler,
    Normalizer,
    RobustScaler,
    StandardScaler,
)

import pandas as pd

from relevanceai.dataset.write.write import Write


class Scaler(Write):
    def scale(
        self,
        fields: List[Any],
        output_field=None,
        scaler: Any = None,
        number_of_documents: Union[int, None] = None,
        show_progress_bar: bool = True,
    ):

        filters = [
            {
                "field": field,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": "",
            }
            for field in fields
        ]

        print("Getting documents")
        if number_of_documents is None:
            documents = self.get_all_documents(
                select_fields=fields,
                include_vector=True,
                filters=filters,
                show_progress_bar=show_progress_bar,
            )
        else:
            documents = self.get_documents(
                select_fields=fields,
                number_of_documents=number_of_documents,
                include_vector=True,
                filters=filters,
                show_progress_bar=show_progress_bar,
            )

        schema = self.schema
        vector_fields = [field for field in schema if "_vector_" in field]
        distributed_fields = []

        print("Scaling fields")
        data: Dict[str, Any] = {"_id": []}
        for document in documents:
            for field in document.keys():
                if field != "_id":
                    value = document[field]

                    if field in vector_fields:
                        if field not in distributed_fields:
                            distributed_fields.append(field)

                        n_dims = len(value)
                        for dim in range(n_dims):
                            vector_field_dim = f"{field}-{dim}"
                            if vector_field_dim not in data:
                                data[vector_field_dim] = []
                            data[vector_field_dim].append(value[dim])
                    else:
                        if field not in data:
                            data[field] = []
                        data[field].append(value)
                else:
                    data["_id"].append(document[field])

        df = pd.DataFrame(data)

        if scaler is None:
            scaler = RobustScaler()

        non_id_cols = [col for col in df.columns if col != "_id"]
        scaled = scaler.fit_transform(df[non_id_cols].values)
        for index, column in enumerate(non_id_cols):
            df[column] = scaled[:, index].tolist()

        documents = df.to_dict("records")

        for document in documents:
            for field in distributed_fields:
                dims = schema[field]["vector"]
                document[field if output_field is None else output_field] = [
                    document[f"{field}-{n}"] for n in range(dims)
                ]
                for n in range(dims):
                    document.pop(f"{field}-{n}")

        print("Updating fields")
        self.update_documents(
            self.dataset_id, documents, show_progress_bar=show_progress_bar
        )



import random

from typing import NamedTuple

from dataclasses import dataclass

from relevanceai.utils import make_id


@dataclass
class DataclassDocument:
    _id: str = None
    value1: float = random.random()
    value2: float = random.random()


def dataclass_document() -> NamedTuple:
    document = DataclassDocument()
    document._id = make_id(document)
    return document



"""Services class
"""
from typing import Optional

from relevanceai.utils.base import _Base

from relevanceai._api.endpoints.services.encoders import EncodersClient
from relevanceai._api.endpoints.services.cluster import ClusterClient
from relevanceai._api.endpoints.services.search import SearchClient
from relevanceai._api.endpoints.services.aggregate import AggregateClient
from relevanceai._api.endpoints.services.recommend import RecommendClient
from relevanceai._api.endpoints.services.tagger import TaggerClient
from relevanceai._api.endpoints.services.prediction import PredictionClient
from relevanceai._api.endpoints.services.wordclouds import WordcloudsClient


class ServicesClient(_Base):
    def __init__(self, credentials):
        self.encoders = EncodersClient(credentials)
        self.cluster = ClusterClient(credentials)
        self.search = SearchClient(credentials)
        self.aggregate = AggregateClient(credentials)
        self.recommend = RecommendClient(credentials)
        self.tagger = TaggerClient(credentials)
        self.prediction = PredictionClient(credentials)
        self.wordclouds = WordcloudsClient(credentials)
        super().__init__(credentials)

    def document_diff(
        self,
        doc: dict,
        documents_to_compare: list,
        difference_fields: Optional[list] = None,
    ):
        """
        Find differences between documents

        Parameters
        ----------
        doc: dict
            Main document to compare other documents against.
        documents_to_compare: list
            Other documents to compare against the main document.
        difference_fields: list
            Fields to compare. Defaults to [], which compares all fields.

        """
        difference_fields = [] if difference_fields is None else difference_fields

        return self.make_http_request(
            endpoint=f"/services/document_diff",
            method="POST",
            parameters={
                "doc": doc,
                "documents_to_compare": documents_to_compare,
                "difference_fields": difference_fields,
            },
        )



"""
Checks for available integrations in the Python module.
"""

import importlib


def is_faiss_available():
    return importlib.util.find_spec("faiss") is not None


def is_scipy_available():
    return importlib.util.find_spec("scipy") is not None


def is_sklearn_available():
    if importlib.util.find_spec("sklearn") is None:
        return False
    return is_scipy_available() and importlib.util.find_spec("sklearn.metrics")


def is_hdbscan_available():
    return importlib.util.find_spec("hdbscan") is not None


def is_pandas_available():
    return importlib.util.find_spec("pandas") is not None


def is_plotly_available():
    return importlib.util.find_spec("plotly") is not None


def is_transformers_available():
    return importlib.util.find_spec("transformers") is not None



from doc_utils.doc_utils import DocUtils
from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base
from relevanceai.operations.cluster.constants import NEAREST_NEIGHBOURS

doc_utils = DocUtils()


class NearestNeighbours(_Base, DocUtils):
    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    @staticmethod
    def get_nearest_neighbours(
        documents: list,
        vector: list,
        vector_field: str,
        distance_measure_mode: NEAREST_NEIGHBOURS = "cosine",
        callable_distance=None,
        score_field: str = "nearest_neighbour_distance",
    ):
        import scipy.spatial.distance as spatial_distance

        if callable_distance:
            sort_key = [
                callable_distance(i, vector)
                for i in doc_utils.get_field_across_documents(vector_field, documents)
            ]
            reverse = False

        elif distance_measure_mode == "cosine":
            # TOOD: multiprocess this cos this is weak
            sort_key = [
                1 - spatial_distance.cosine(i, vector)
                for i in doc_utils.get_field_across_documents(vector_field, documents)
            ]
            reverse = True

        elif distance_measure_mode == "l2":
            sort_key = [
                spatial_distance.euclidean(i, vector)
                for i in doc_utils.get_field_across_documents(vector_field, documents)
            ]
            reverse = False

        else:
            raise ValueError("Need valid distance measure mode or callable distance")

        doc_utils.set_field_across_documents(score_field, sort_key, documents)

        return sorted(documents, reverse=reverse, key=lambda x: x[score_field])



"""Code for migrating datasets
"""
from typing import Optional


def migrate_dataset(
    old_token: str,
    new_token: str,
    dataset_id: str,
    new_dataset_id: Optional[str] = None,
    chunksize: int = 100,
    filters: list = None,
):
    """
    Migrate dataset

    Args
    ---------
        old_token (str): _description_
        new_token (str): _description_
        dataset_id (str): _description_
        new_dataset_id (Optional[str], optional): _description_. Defaults to None.
        chunksize (int, optional): _description_. Defaults to 20.

    Example
    ---------

    .. code-block::

        from relevanceai.migration import migrate
        migrate_dataset(
            old_token="...",
            new_token="...",
            dataset_id="sample_dataset",
            new_dataset_id="new_sample_dataset")

    """
    from relevanceai import Client
    from relevanceai.utils.logger import FileLogger

    filters = filters if filters is not None else []

    if new_dataset_id is None:
        new_dataset_id = dataset_id

    client = Client(token=old_token)
    ds = client.Dataset(dataset_id)
    with FileLogger():
        docs = ds.get_documents(
            number_of_documents=chunksize, include_cursor=True, filters=filters
        )
        while len(docs["documents"]) > 0:
            new_client = Client(token=new_token)
            new_ds = new_client.Dataset(new_dataset_id)
            for d in docs["documents"]:
                # backwards compatibility for clusters
                if "_clusters_" in d:
                    d["_cluster_"] = d.pop("_clusters_")
            new_ds.upsert_documents(docs["documents"])
            # we need to reset the config
            client = Client(token=old_token)
            docs = ds.get_documents(
                number_of_documents=chunksize,
                include_cursor=True,
                cursor=docs["cursor"],
                filters=filters,
            )
    print("Finished migrating.")



import pytest

from relevanceai.client import Client
from relevanceai.dataset import Dataset
from relevanceai.operations.cluster import ClusterOps


@pytest.mark.skip(reason="KMeansModel missing")
def test_dataset_api_kmeans_centroids_properties(
    test_client: Client, test_dataset: Dataset
):

    alias: str = "test_alias"
    vector_field: str = "sample_1_vector_"

    from relevanceai.operations.cluster.models.kmeans import KMeansModel

    model = KMeansModel()

    clusterer: ClusterOps = test_client.ClusterOps(model=model, alias=alias)
    clusterer.fit(dataset=test_df, vector_fields=[vector_field])

    assert f"_cluster_.{vector_field}.{alias}" in test_dataset.schema

    # TODO: see why centroids fail
    centroids = clusterer.list_closest_to_center()
    assert len(centroids) > 0



import numpy as np

from relevanceai.utils import make_id


def numpy_document():
    document = {
        "sample_1_numpy": np.random.randint(5, size=1)[0],
        "sample_2_numpy": np.random.rand(3, 2),
        "sample_3_numpy": np.nan,
    }
    document["_id"] = make_id(document)
    return document



from relevanceai.dataset import Dataset


# Skip cluster tests
# def test_cluster(test_dataset: Dataset):
#     from relevanceai.operations.cluster.models.kmeans import KMeansModel

#     vector_field = "sample_1_vector_"
#     alias = "test_alias"

#     model = KMeansModel()

#     test_dataset.cluster(model=model, alias=alias, vector_fields=[vector_field])
#     assert f"_cluster_.{vector_field}.{alias}" in test_dataset.schema


# def test_centroids(test_clustered_df: Dataset):
#     closest = test_clustered_df.centroids(["sample_1_vector_"], "kmeans-10").closest()
#     assert "results" in closest
#     assert all("cluster" in cluster for cluster in list(closest["results"]))

#     furthest = test_clustered_df.centroids(["sample_1_vector_"], "kmeans-10").furthest()
#     assert "results" in furthest
#     assert all("cluster" in cluster for cluster in list(furthest["results"]))


def test_health(test_dataset: Dataset):
    import pandas as pd

    dataframe_output = test_dataset.health(output_format="dataframe")
    assert type(dataframe_output) == pd.DataFrame
    json_output = test_dataset.health(output_format="json")
    assert type(json_output) == dict



# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
import pandas as pd

from typing import Optional, Union

from relevanceai.client.helpers import Credentials
from relevanceai.dataset.series import Series
from relevanceai.utils.decorators.version import added
from relevanceai.utils.decorators.analytics import track
from relevanceai._api import APIClient
from relevanceai._api.endpoints.services.cluster import ClusterClient


class Statistics(APIClient):
    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
        **kwargs,
    ):
        self.credentials = credentials
        self.dataset_id = dataset_id
        super().__init__(
            dataset_id=self.dataset_id,
            credentials=credentials,
            **kwargs,
        )

    @track
    def value_counts(self, field: str):
        """
        Return a Series containing counts of unique values.

        Parameters
        ----------
        field: str
            dataset field to which to do value counts on

        Returns
        -------
        Series

        Example
        -----------------
        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            field = "sample_field"
            value_counts_df = df.value_counts(field)

        """
        return Series(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            field=field,
        ).value_counts()

    @track
    def describe(self, return_type="pandas") -> dict:
        """
        Descriptive statistics include those that summarize the central tendency
        dispersion and shape of a dataset's distribution, excluding NaN values.


        Example
        -----------------
        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            field = "sample_field"
            df.describe() # returns pandas dataframe of stats
            df.describe(return_type='dict') # return raw json stats

        """
        facets = self.datasets.facets(self.dataset_id)
        if return_type == "pandas":
            schema = self.datasets.schema(self.dataset_id)
            dataframe = {
                col: facets["results"][col]
                for col in schema
                if col in facets["results"] and isinstance(facets["results"][col], dict)
            }
            dataframe = pd.DataFrame(dataframe)
            return dataframe
        elif return_type == "dict":
            return facets
        else:
            raise ValueError("invalid return_type, should be `dict` or `pandas`")

    @added(version="1.2.2")
    def corr(
        self,
        X: str,
        Y: str,
        vector_field: str,
        alias: str,
        groupby: Optional[str] = None,
        fontsize: int = 16,
    ):
        """
        Returns the Pearson correlation between two fields.

        Parameters
        ----------
        X: str
            A dataset field

        Y: str
            The other dataset field

        vector_field: str
            The vector field over which the clustering has been performed

        alias: str
            The alias of the clustering

        groupby: Optional[str]
            A field to group the correlations over

        fontsize: int
            The font size of the values in the image
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        # todo: how to cover cases when fields are in schema but not "calculable" fields like clusters and deployables
        cclient = ClusterClient(self.credentials)
        groupby_agg = (
            []
            if groupby is None
            else [{"name": groupby, "field": groupby, "agg": "correlation"}]
        )
        res = cclient.aggregate(
            dataset_id=self.dataset_id,
            vector_fields=[vector_field],
            metrics=[{"name": "correlation", "fields": [X, Y], "agg": "correlation"}],
            groupby=groupby_agg,
            alias=alias,
        )["results"]

        clusters = sorted(res.keys())

        if groupby is None:
            categories = ["cluster"]
        else:
            series = Series(
                credentials=self.credentials,
                dataset_id=self.dataset_id,
                field=groupby,
            ).all(show_progress_bar=False)

            categories = sorted(
                pd.Series(map(lambda _: _[groupby], series)).drop_duplicates()
            )

        # Use a pandas DataFrame for easy indexing
        dataframe = pd.DataFrame(data=[], columns=clusters, index=categories)

        for cluster, values in res.items():
            for value in values:
                correlation_value = value["correlation"][X][Y]
                category = value.get(groupby, "cluster")
                dataframe.at[category, cluster] = correlation_value

        # Only needed pandas DataFrame for indexing, now convert to numpy
        # ndarray for convenience.
        data = dataframe.to_numpy(dtype=float)

        fig, ax = plt.subplots(figsize=(2 * len(clusters), 2 * len(categories)))
        cmap = plt.get_cmap("coolwarm_r")
        cmap.set_bad("k")  # if np.nan, set imshow cell to black
        im = ax.imshow(data, norm=mpl.colors.Normalize(vmin=-1, vmax=1), cmap=cmap)

        ax.set_xticks(
            range(data.shape[1]),
            labels=clusters,
            rotation=-30,
            rotation_mode="anchor",
            ha="right",
            fontsize=fontsize + 1,
        )
        ax.set_yticks(range(data.shape[0]), labels=categories, fontsize=fontsize + 1)
        ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)

        ax.spines[:].set_visible(False)
        ax.set_xticks([n - 0.5 for n in range(data.shape[1] + 1)], minor=True)
        ax.set_yticks([n - 0.5 for n in range(data.shape[0] + 1)], minor=True)
        ax.grid(which="minor", color="w", linestyle="-", linewidth=3)
        ax.tick_params(which="minor", bottom=False, left=False)

        for column, _ in enumerate(clusters):
            for row, _ in enumerate(categories):
                # Ensure that the negative sign doesn't offset the text by
                # offseting positive values.
                if data[row][column] < 0:
                    text = f"{data[row][column]:.2f}"
                else:
                    text = f" {data[row][column]:.2f}"
                im.axes.text(
                    column,
                    row,
                    text,
                    dict(horizontalalignment="center", verticalalignment="center"),
                    fontsize=fontsize,
                )

        fig.tight_layout()
        plt.show()

    def health(self, output_format="dataframe") -> Union[pd.DataFrame, dict]:
        """
        Gives you a summary of the health of your vectors, e.g. how many documents with vectors are missing, how many documents with zero vectors

        Parameters
        ----------

        output_format: str
            The format of the output. Must either be "dataframe" or "json".

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")
            df.health

        """
        results = self.datasets.monitor.health(self.dataset_id)
        if output_format == "dataframe":
            return pd.DataFrame(results).T
        elif output_format == "json":
            return results
        else:
            raise ValueError('\'output_format\' must either be "dataframe" or "json"')

    @track
    def aggregate(
        self,
        groupby: Optional[list] = None,
        metrics: Optional[list] = None,
        filters: Optional[list] = None,
        # sort: list = [],
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        alias: str = "default",
    ):
        return self.services.aggregate.aggregate(
            dataset_id=self.dataset_id,
            groupby=[] if groupby is None else groupby,
            metrics=[] if metrics is None else metrics,
            filters=[] if filters is None else filters,
            page_size=page_size,
            page=page,
            asc=asc,
            flatten=flatten,
            alias=alias,
            # sort=sort
        )

    def facets(
        self,
        fields: Optional[list] = [],
        date_interval: str = "monthly",
        page_size: int = 5,
        page: int = 1,
        asc: bool = False,
    ):
        """
        Get a summary of fields - such as most common, their min/max, etc.

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            from relevanceai.datasets import mock_documents
            documents = mock_documents(100)
            ds = client.Dataset("mock_documents")
            ds.upsert_documents(documents)
            ds.facets(["sample_1_value"])
        """
        return self.datasets.facets(
            dataset_id=self.dataset_id,
            fields=fields,
            date_interval=date_interval,
            page_size=page_size,
            page=page,
            asc=asc,
        )



import os
import json
import warnings
import getpass

import numpy as np

from typing import Union, Callable, Optional, List, Dict, Any

from doc_utils import DocUtils
from relevanceai._api import APIClient

from relevanceai.operations.cluster.base import (
    ClusterBase,
    CentroidClusterBase,
    BatchClusterBase,
    HDBSCANClusterBase,
    SklearnCentroidBase,
)

from relevanceai.utils.integration_checks import (
    is_sklearn_available,
    is_hdbscan_available,
)

from relevanceai.reports.cluster.evaluate import ClusterEvaluate


class _ClusterOps(APIClient, DocUtils):

    alias: str
    dataset_id: str
    vector_fields: List[Any]

    def _assign_sklearn_model(self, model):
        from sklearn.cluster import (
            KMeans,
            MiniBatchKMeans,
            DBSCAN,
            Birch,
            SpectralClustering,
            OPTICS,
            AgglomerativeClustering,
            AffinityPropagation,
            MeanShift,
            FeatureAgglomeration,
        )

        POSSIBLE_MODELS = [
            SpectralClustering,
            Birch,
            DBSCAN,
            OPTICS,
            AgglomerativeClustering,
            AffinityPropagation,
            MeanShift,
            FeatureAgglomeration,
        ]

        if is_hdbscan_available():
            import hdbscan

            if hasattr(hdbscan, "HDBSCAN"):
                POSSIBLE_MODELS.append(hdbscan.HDBSCAN)

        if model.__class__ == KMeans:

            class CentroidClusterModel(CentroidClusterBase):
                def __init__(self, model):
                    self.model: Union[KMeans, MiniBatchKMeans] = model

                def fit_predict(self, X):
                    return self.model.fit_predict(X)

                def get_centers(self):
                    return self.model.cluster_centers_

            new_model = CentroidClusterModel(model)
            return new_model

        elif model.__class__ == MiniBatchKMeans:

            class BatchCentroidClusterModel(CentroidClusterBase, BatchClusterBase):
                def __init__(self, model):
                    self.model: MiniBatchKMeans = model

                def partial_fit(self, X):
                    return self.model.partial_fit(X)

                def predict(self, X):
                    return self.model.predict(X)

                def get_centers(self):
                    return self.model.cluster_centers_

            new_model = BatchCentroidClusterModel(model)
            return new_model

        elif isinstance(model, tuple(POSSIBLE_MODELS)):
            if "sklearn" in str(type(model)).lower():
                new_model = SklearnCentroidBase(model)

            elif "hdbscan" in str(type(model)).lower():
                new_model = HDBSCANClusterBase(model)

            return new_model

        elif hasattr(model, "fit_documents"):
            return model

        elif hasattr(model, "fit_predict"):
            data = {"fit_predict": model.fit_predict, "metadata": model.__dict__}
            ClusterModel = type("ClusterBase", (ClusterBase,), data)
            return ClusterModel()

        elif hasattr(model, "fit_transform"):
            data = {"fit_predict": model.fit_transform, "metadata": model.__dict__}
            ClusterModel = type("ClusterBase", (ClusterBase,), data)
            return ClusterModel()

    def _assign_model(self, model):

        if (is_sklearn_available() or is_hdbscan_available()) and (
            "sklearn" in str(type(model)).lower()
            or "hdbscan" in str(type(model)).lower()
        ):
            model = self._assign_sklearn_model(model)
            if model is not None:
                return model

        if isinstance(model, ClusterBase):
            return model

        elif hasattr(model, "fit_documents"):
            return model

        elif hasattr(model, "fit_predict"):
            data = {"fit_predict": model.fit_predict, "metadata": model.__dict__}
            ClusterModel = type("ClusterBase", (ClusterBase,), data)
            return ClusterModel()

        elif model is None:
            return model

        raise TypeError("Model should be inherited from ClusterBase.")

    def _token_to_auth(self, token=None):
        SIGNUP_URL = "https://cloud.relevance.ai/sdk/api"

        if os.path.exists(self._cred_fn):
            credentials = self._read_credentials()
            return credentials

        elif token:
            return self._process_token(token)

        else:
            print(f"Activation token (you can find it here: {SIGNUP_URL} )")
            if not token:
                token = getpass.getpass(f"Activation token:")
            return self._process_token(token)

    def _process_token(self, token: str):
        split_token = token.split(":")
        project = split_token[0]
        api_key = split_token[1]
        if len(split_token) > 2:
            region = split_token[3]
            base_url = self._region_to_url(region)

            if len(split_token) > 3:
                firebase_uid = split_token[4]
                return self._write_credentials(
                    project=project,
                    api_key=api_key,
                    base_url=base_url,
                    firebase_uid=firebase_uid,
                )

            else:
                return self._write_credentials(
                    project=project, api_key=api_key, base_url=base_url
                )

        else:
            return self._write_credentials(project=project, api_key=api_key)

    def _read_credentials(self):
        return json.load(open(self._cred_fn))

    def _write_credentials(self, **kwargs):
        print(
            f"Saving credentials to {self._cred_fn}. Remember to delete this file if you do not want credentials saved."
        )
        json.dump(
            kwargs,
            open(self._cred_fn, "w"),
        )
        return kwargs

    def _insert_centroid_documents(self):
        if hasattr(self.model, "get_centroid_documents"):
            print("Inserting centroid documents...")
            centers = self.get_centroid_documents()

            # Change centroids insertion
            results = self.services.cluster.centroids.insert(
                dataset_id=self.dataset_id,
                cluster_centers=centers,
                vector_fields=self.vector_fields,
                alias=self.alias,
            )
            self.logger.info(results)

        return

    def _check_for_dataset_id(self):
        if not hasattr(self, "dataset_id"):
            raise ValueError(
                "You are missing a dataset ID. Please set using the argument dataset_id='...'."
            )

    def _concat_vectors_from_list(self, list_of_vectors: list):
        """Concatenate 2 vectors together in a pairwise fashion"""
        return [np.concatenate(x) for x in list_of_vectors]

    def _get_vectors_from_documents(
        self, vector_fields: List[Any], documents: List[Dict]
    ):
        if len(vector_fields) == 1:
            # filtering out entries not containing the specified vector
            documents = list(filter(DocUtils.list_doc_fields, documents))
            vectors = self.get_field_across_documents(
                vector_fields[0], documents, missing_treatment="skip"
            )
        else:
            # In multifield clusering, we get all the vectors in each document
            # (skip if they are missing any of the vectors)
            # Then run clustering on the result
            documents = list(self.filter_docs_for_fields(vector_fields, documents))
            all_vectors = self.get_fields_across_documents(
                vector_fields, documents, missing_treatment="skip_if_any_missing"
            )
            # Store the vector field lengths to de-concatenate them later
            self._vector_field_length: dict = {}
            prev_vf = 0
            for i, vf in enumerate(self.vector_fields):
                self._vector_field_length[vf] = {}
                self._vector_field_length[vf]["start"] = prev_vf
                end_vf = prev_vf + len(all_vectors[0][i])
                self._vector_field_length[vf]["end"] = end_vf
                # Update the ending
                prev_vf = end_vf

            # Store the vector lengths
            vectors = self._concat_vectors_from_list(all_vectors)

        return vectors

    def _chunk_dataset(
        self,
        dataset_id: str,
        select_fields: Optional[list] = None,
        chunksize: int = 100,
        filters: Optional[list] = None,
    ):
        """Utility function for chunking a dataset"""
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters

        docs = self._get_documents(
            dataset_id=self.dataset_id,
            include_cursor=True,
            number_of_documents=chunksize,
            select_fields=select_fields,
            filters=filters,
        )

        while len(docs["documents"]) > 0:
            yield docs["documents"]
            docs = self._get_documents(
                dataset_id=self.dataset_id,
                cursor=docs["cursor"],
                include_cursor=True,
                select_fields=select_fields,
                number_of_documents=chunksize,
                filters=filters,
            )

    def _get_parent_cluster_values(
        self, vector_fields: list, alias: str, documents
    ) -> list:
        field = ".".join([self.cluster_field, ".".join(sorted(vector_fields)), alias])
        return self.get_field_across_documents(
            field, documents, missing_treatment="skip"
        )

    @staticmethod
    def _calculate_silhouette_grade(vectors, cluster_labels):
        from relevanceai.reports.cluster.grading import (
            get_silhouette_grade,
        )
        from sklearn.metrics import silhouette_samples

        score = silhouette_samples(vectors, cluster_labels, metric="euclidean").mean()
        grade = get_silhouette_grade(score)

        print("---------------------------")
        print(f"Grade: {grade}")
        print(f"Mean Silhouette Score: {score}")
        print("---------------------------")

    def _get_cluster_field_name(self, alias: str = None):
        if alias is None:
            alias = self.alias
        if isinstance(self.vector_fields, list):
            set_cluster_field = (
                f"{self.cluster_field}.{'.'.join(self.vector_fields)}.{alias}"
            )
        elif isinstance(self.vector_fields, str):
            set_cluster_field = f"{self.cluster_field}.{self.vector_fields}.{alias}"
        return set_cluster_field

    def _set_cluster_labels_across_documents(self, cluster_labels, documents):
        set_cluster_field = self._get_cluster_field_name()
        self.set_field_across_documents(set_cluster_field, cluster_labels, documents)

    def _label_cluster(self, label: Union[int, str]):
        if not isinstance(label, str):
            return "cluster-" + str(label)
        return str(label)

    def _label_subcluster(self, label: Union[int, str], prev_cluster_label) -> str:
        return prev_cluster_label + "-" + str(label)

    def _label_clusters(self, labels):
        return [self._label_cluster(x) for x in labels]

    def _label_subclusters(self, labels: List[str], prev_cluster_labels: List[str]):
        return [
            self._label_subcluster(label, prev_cluster_label)
            for label, prev_cluster_label in zip(labels, prev_cluster_labels)
        ]

    def _operate(self, cluster_id: str, field: str, output: dict, func: Callable):
        """
        Internal function for operations
        """
        cluster_field = self._get_cluster_field_name()
        # TODO; change this to fetch all documents
        documents = self.datasets.documents.get_where(
            self.dataset_id,
            filters=[
                {
                    "field": cluster_field,
                    "filter_type": "exact_match",
                    "condition": "==",
                    "condition_value": cluster_id,
                },
                {
                    "field": field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                },
            ],
            select_fields=[field, cluster_field],
            page_size=9999,
        )
        # get the field across each
        arr = self.get_field_across_documents(field, documents["documents"])
        output[cluster_id] = func(arr)

    def _get_filter_for_cluster(self, cluster_id):
        cluster_field = self._get_cluster_field_name()
        filters = [
            {
                "field": cluster_field,
                "filter_type": "exact_match",
                "condition": "==",
                "condition_value": cluster_id,
            }
        ]
        return filters


# class _ClusterOpsShow(Read):
#     def __init__(self, df, is_image_field: bool):
#         self.df = df
#         if is_image_field:
#             self.text_fields = []
#             self.image_fields = df.columns.tolist()
#         else:
#             self.text_fields = df.columns.tolist()
#             self.image_fields = []

#     def _repr_html_(self):
#         try:
#             documents = self.df.to_dict(orient="records")
#             return self._show_json(documents, return_html=True)
#         except Exception as e:
#             warnings.warn(
#                 "Displaying using pandas. To get image functionality please install RelevanceAI[notebook]. "
#                 + str(e)
#             )
#             return self.df._repr_html_()

#     def _show_json(self, documents, **kw):
#         from jsonshower import show_json

#         return show_json(
#             documents,
#             text_fields=self.text_fields,
#             image_fields=self.image_fields,
#             **kw,
#         )



"""
All admin-related tasks.
"""
from typing import Optional

from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base


class AdminClient(_Base):
    def __init__(self, credentials: Credentials):
        self.project = credentials.project
        self.api_key = credentials.api_key

        super().__init__(credentials)

    def request_read_api_key(self, read_username: str):
        """Creates a read only key for your project. Make sure to save the api key somewhere safe. When doing a search the admin username should still be used.

        Parameters
        ----------
        read_username:
            Read-only project
        """
        return self.make_http_request(
            "/admin/request_read_api_key",
            method="POST",
            parameters={"read_username": read_username},
        )

    def copy_foreign_dataset(
        self,
        dataset_id: str,
        source_dataset_id: str,
        source_project: Optional[str],
        source_api_key: Optional[str],
        project: str = None,
        api_key: str = None,
    ):
        """Copy a dataset from another user's projects into your project.

        Example
        -----------

        .. code-block::

            client = Client()
            client.admin.send_dataset(
                dataset_id="research",
                receiver_project="...",
                receiver_api_key="..."
            )

        Parameters
        -------------

        dataset_id:
            The dataset to copy
        source_dataset_id:
            The original dataset
        source_project:
            The original project to copy from
        source_api_key:
            The original API key of the project
        project:
            The original project
        api_key:
            The original API key
        """
        return self.make_http_request(
            "/admin/copy_foreign_dataset",
            method="POST",
            parameters={
                "project": self.project if project is None else project,
                "api_key": self.api_key if api_key is None else api_key,
                "dataset_id": dataset_id,
                "source_dataset_id": source_dataset_id,
                "source_project": source_project,
                "source_api_key": source_api_key,
                # "filters": filters
            },
        )

    def send_dataset(
        self,
        dataset_id: str,
        receiver_project: str,
        receiver_api_key: str,
    ):
        """
        Send an individual a dataset.

        Example
        --------
        >>> client = Client()
        >>> client.admin.send_dataset(
            dataset_id="research",
            receiver_project="...",
            receiver_api_key="..."
        )

        Parameters
        -----------

        dataset_id: str
            The name of the dataset
        receiver_project: str
            The project name that will receive the dataset
        receiver_api_key: str
            The project API key that will receive the dataset

        """
        return self.make_http_request(
            "/admin/copy_foreign_dataset",
            method="POST",
            parameters={
                "project": receiver_project,
                "api_key": receiver_api_key,
                "dataset_id": dataset_id,
                "source_dataset_id": dataset_id,
                "source_project": self.project,
                "source_api_key": self.api_key,
                # "filters": filters
            },
        )

    def receive_dataset(
        self,
        dataset_id: str,
        sender_project: str,
        sender_api_key: str,
    ):
        """
        Receive an individual a dataset.

        Example
        --------
        >>> client = Client()
        >>> client.admin.receive_dataset(
            dataset_id="research",
            sender_project="...",
            sender_api_key="..."
        )

        Parameters
        -----------

        dataset_id: str
            The name of the dataset
        sender_project: str
            The project name that will send the dataset
        sender_api_key: str
            The project API key that will send the dataset

        """
        return self.make_http_request(
            "/admin/copy_foreign_dataset",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "source_dataset_id": dataset_id,
                "project": self.project,
                "api_key": self.api_key,
                "source_project": sender_project,
                "source_api_key": sender_api_key,
                # "filters": filters
            },
        )

    def _ping(self):
        return self.make_http_request("/admin/ping", method="GET")



"""Utility functions for useful printing formats.
"""


def print_hashtag_message(message) -> None:
    loud_format = "###" + len(message) * "#" + "###"
    message_format = "## " + message + " ##"

    print(loud_format)
    print(message_format)
    print(loud_format)


class COLOR:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def print_bold_message(message):
    print(COLOR.BOLD + message + COLOR.END)



import pandas as pd

from relevanceai.dataset.read import Read
from relevanceai.utils.decorators.analytics import track


class CSVExport(Read):
    @track
    def to_csv(self, filename: str, **kwargs):
        """
        Download a dataset from Relevance AI to a local .csv file

        Parameters
        ----------
        filename: str
            path to downloaded .csv file
        kwargs: Optional
            see client.get_all_documents() for extra args

        Example
        -------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            csv_fname = "path/to/csv/file.csv"
            df.to_csv(csv_fname)
        """
        documents = self.get_all_documents(**kwargs)
        df = pd.DataFrame(documents)
        df.to_csv(filename)



import getpass

from dataclasses import dataclass

from relevanceai.constants.messages import Messages
from relevanceai.constants import (
    SIGNUP_URL,
    AUSTRALIA_URL,
    WIDER_URL,
    OLD_AUSTRALIA_EAST,
    DEV_URL,
)
from relevanceai.constants.errors import (
    APIKeyNotFoundError,
    FireBaseUIDNotFoundError,
    ProjectNotFoundError,
    RegionNotFoundError,
    TokenNotFoundError,
)


def region_to_url(region: str):
    """
    Convert region to URL
    """
    if "dev" in region:
        actual_region = region.replace("dev-", "")
        url = DEV_URL.format(actual_region)
    elif region == OLD_AUSTRALIA_EAST:
        url = AUSTRALIA_URL
    else:
        url = WIDER_URL.format(region)
    return url


@dataclass(frozen=True)
class Credentials:
    """
    A convenience store of relevant credentials.
    """

    __slots__ = (
        "token",
        "project",
        "api_key",
        "region",
        "firebase_uid",
    )
    token: str
    project: str
    api_key: str
    region: str
    firebase_uid: str

    def split_token(self):
        return self.token.split(":")


def process_token(token: str):
    if not token:
        raise TokenNotFoundError

    # A token takes the format project:api_key:region:firebase_uid
    project, api_key, *other_credentials = token.split(":")

    if not project:
        raise ProjectNotFoundError

    if not api_key:
        raise APIKeyNotFoundError

    if len(other_credentials) == 0:
        # Assume that region is missing as that is the next credential
        raise RegionNotFoundError
    elif len(other_credentials) == 1:
        # Assume region exists but firebase_uid, the next credential, does not
        raise FireBaseUIDNotFoundError
    else:
        # Two or more other credentials is correct
        region, firebase_uid, *additional_credentials = other_credentials

    return Credentials(token, project, api_key, region, firebase_uid)


def auth():
    print(Messages.TOKEN_MESSAGE.format(SIGNUP_URL))
    token = getpass.getpass(f"Activation Token: ")
    return token



import warnings

from functools import wraps

from relevanceai.constants.messages import Messages


def beta(f):
    f.__doc__ = Messages.BETA_DOCSTRING.format(f.__doc__)

    @wraps(f)
    def wrapper(*args, **kwds):
        return f(*args, **kwds)

    return wrapper


def added(version):
    def _version(f):
        f.__doc__ = Messages.ADDED_DOCSTRING.format(version)

        @wraps(f)
        def wrapper(*args, **kwds):
            return f(*args, **kwds)

        return wrapper

    return _version


def deprecated(version: str, message: str = ""):
    def _version(f):
        f.__doc__ = Messages.DEPRECEATED_DOCSTRING.format(version)

        @wraps(f)
        def wrapper(*args, **kwds):
            warnings.warn(
                Messages.DEPRECEATED.format(version, message), DeprecationWarning
            )
            return f(*args, **kwds)

        return wrapper

    return _version



import pytest


# @pytest.mark.skip(reason="bad API_KEY on old aus east")
def test_get_ecommerce_dataset_subset():
    from relevanceai.utils.datasets import get_ecommerce_3_dataset

    assert len(get_ecommerce_3_dataset(number_of_documents=1000)) == 1000



import pytest

from typing import Dict, List

from relevanceai import Client

from tests.globals.constants import generate_dataset_id


@pytest.fixture(scope="function")
def vector_dataset_id(test_client: Client, vector_documents: List[Dict]):
    test_dataset_id = generate_dataset_id()
    test_client._insert_documents(test_dataset_id, vector_documents)
    yield test_dataset_id
    test_client.datasets.delete(test_dataset_id)



from typing import Optional

from relevanceai.client.helpers import Credentials
from relevanceai.dashboard.dashboard import Dashboard


class Clusters(Dashboard):
    def __init__(self, credentials: Credentials, project: str):
        super().__init__(credentials, project, "cluster")

    @classmethod
    def create_dashboard(
        cls,
        credentials: Credentials,
        dataset_id: str,
        vector_field: str,
        alias: str,
        share: bool = False,
        **configuration
    ):
        # TODO: if there is no _cluster_ field in schema, create it here?
        application = "cluster"
        application_configuration = {
            "collection_name": dataset_id,
            "type": application,
            "deployable_name": dataset_id,
            "project_id": credentials.project,
            application: {
                "alias": alias,
                "vector_field": vector_field,
                **{
                    key: value
                    for key, value in configuration.items()
                    if key not in {"alias", "vector_field"}
                },
            },
        }

        return Dashboard.create_dashboard(
            credentials,
            dataset_id,
            vector_field,
            application,
            share,
            application_configuration,
        )
